from collections import OrderedDict
from itertools import product
from typing import List, Optional, Union

import numpy as np
import scipy.sparse as sparse
from pandas import DataFrame, Series
from sklearn.metrics import confusion_matrix

from ..utils import to_int_label_array

from .lf import LabelingFunction


class LFAnalysis:
    """Run analyses on LFs using label matrix.

    Parameters
    ----------
    L
        Label matrix where L_{i,j} is the label given by the jth LF to the ith
        candidate (using -1 for abstain)
    lfs
        Labeling functions used to generate ``L``

    Raises
    ------
    ValueError
        If number of LFs and number of LF matrix columns differ

    Attributes
    ----------
    L
        See above.
    """

    def __init__(
        self, L: np.ndarray, lfs: Optional[List[LabelingFunction]] = None
    ) -> None:
        self.L = L
        self._L_sparse = sparse.csr_matrix(L + 1)
        self._lf_names = None
        if lfs is not None:
            if len(lfs) != self._L_sparse.shape[1]:
                raise ValueError(
                    f"Number of LFs ({len(lfs)}) and number of "
                    f"LF matrix columns ({self._L_sparse.shape[1]}) are different"
                )
            self._lf_names = [lf.name for lf in lfs]

    def _covered_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled by at least one LF."""
        return np.ravel(np.where(self._L_sparse.sum(axis=1) != 0, 1, 0))

    def _overlapped_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled by more than one LF."""
        return np.where(np.ravel((self._L_sparse != 0).sum(axis=1)) > 1, 1, 0)

    def _conflicted_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled differently by two LFs."""
        m = sparse.diags(np.ravel(self._L_sparse.max(axis=1).todense()))
        return np.ravel(
            np.max(m @ (self._L_sparse != 0) != self._L_sparse, axis=1)
            .astype(int)
            .todense()
        )

    def label_coverage(self) -> float:
        """Compute the fraction of data points with at least one label.

        Returns
        -------
        float
            Fraction of data points with labels

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).label_coverage()
        0.8
        """
        return self._covered_data_points().sum() / self._L_sparse.shape[0]

    def label_overlap(self) -> float:
        """Compute the fraction of data points with at least two (non-abstain) labels.

        Returns
        -------
        float
            Fraction of data points with overlapping labels

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).label_overlap()
        0.6
        """
        return self._overlapped_data_points().sum() / self._L_sparse.shape[0]

    def label_conflict(self) -> float:
        """Compute the fraction of data points with conflicting (non-abstain) labels.

        Returns
        -------
        float
            Fraction of data points with conflicting labels

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).label_conflict()
        0.2
        """
        return self._conflicted_data_points().sum() / self._L_sparse.shape[0]

    def lf_polarities(self) -> List[List[int]]:
        """Infer the polarities of each LF based on evidence in a label matrix.

        Returns
        -------
        List[List[int]]
            Unique output labels for each LF

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).lf_polarities()
        [[0, 1], [0], [0]]
        """
        return [
            sorted(list(set(self._L_sparse[:, i].data - 1)))
            for i in range(self._L_sparse.shape[1])
        ]

    def lf_coverages(self) -> np.ndarray:
        """Compute frac. of examples each LF labels.

        Returns
        -------
        numpy.ndarray
            Fraction of labeled examples for each LF

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).lf_coverages()
        array([0.4, 0.8, 0.4])
        """
        return np.ravel((self._L_sparse != 0).sum(axis=0)) / self._L_sparse.shape[0]

    def lf_overlaps(self, normalize_by_coverage: bool = False) -> np.ndarray:
        """Compute frac. of examples each LF labels that are labeled by another LF.

        An overlapping example is one that at least one other LF returns a
        (non-abstain) label for.

        Note that the maximum possible overlap fraction for an LF is the LF's
        coverage, unless ``normalize_by_coverage=True``, in which case it is 1.

        Parameters
        ----------
        normalize_by_coverage
            Normalize by coverage of the LF, so that it returns the percent of LF labels
            that have overlaps.

        Returns
        -------
        numpy.ndarray
            Fraction of overlapping examples for each LF

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).lf_overlaps()
        array([0.4, 0.6, 0.4])
        >>> LFAnalysis(L).lf_overlaps(normalize_by_coverage=True)
        array([1.  , 0.75, 1.  ])
        """
        overlaps = (
            (self._L_sparse != 0).T
            @ self._overlapped_data_points()
            / self._L_sparse.shape[0]
        )
        if normalize_by_coverage:
            overlaps /= self.lf_coverages()
        return np.nan_to_num(overlaps)

    def lf_conflicts(self, normalize_by_overlaps: bool = False) -> np.ndarray:
        """Compute frac. of examples each LF labels and labeled differently by another LF.

        A conflicting example is one that at least one other LF returns a
        different (non-abstain) label for.

        Note that the maximum possible conflict fraction for an LF is the LF's
        overlaps fraction, unless ``normalize_by_overlaps=True``, in which case it is 1.

        Parameters
        ----------
        normalize_by_overlaps
            Normalize by overlaps of the LF, so that it returns the percent of LF
            overlaps that have conflicts.

        Returns
        -------
        numpy.ndarray
            Fraction of conflicting examples for each LF

        Example
        -------
        >>> L = np.array([
        ...     [-1, 0, 0],
        ...     [-1, -1, -1],
        ...     [1, 0, -1],
        ...     [-1, 0, -1],
        ...     [0, 0, 0],
        ... ])
        >>> LFAnalysis(L).lf_conflicts()
        array([0.2, 0.2, 0. ])
        >>> LFAnalysis(L).lf_conflicts(normalize_by_overlaps=True)
        array([0.5       , 0.33333333, 0.        ])
        """
        conflicts = (
            (self._L_sparse != 0).T
            @ self._conflicted_data_points()
            / self._L_sparse.shape[0]
        )
        if normalize_by_overlaps:
            conflicts /= self.lf_overlaps()
        return np.nan_to_num(conflicts)

    def lf_empirical_accuracies(self, Y: np.ndarray) -> np.ndarray:
        """Compute empirical accuracy against a set of labels Y for each LF.

        Usually, Y represents development set labels.

        Parameters
        ----------
        Y
            [n] or [n, 1] np.ndarray of gold labels

        Returns
        -------
        numpy.ndarray
            Empirical accuracies for each LF
        """
        Y = to_int_label_array(Y)
        X = np.where(
            self.L == -1,
            0,
            np.where(self.L == np.vstack([Y] * self.L.shape[1]).T, 1, -1),
        )
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.nan_to_num(0.5 * (X.sum(axis=0) / (self.L != -1).sum(axis=0) + 1))

    def lf_empirical_probs(self, Y: np.ndarray, k: int) -> np.ndarray:
        """Estimate conditional probability tables for each LF.

        Computes conditional probability tables, P(L | Y), for each LF using
        the provided true labels Y.

        Parameters
        ----------
        Y
            The n-dim array of true labels in {1,...,k}
        k
            The cardinality i.e. number of classes

        Returns
        -------
        np.ndarray
            An m x (k+1) x k np.ndarray representing the m (k+1) x k conditional probability
            tables P_i, where P_i[l,y] represents P(LF_i = l | Y = y) empirically calculated
        """
        n, m = self.L.shape

        Y = to_int_label_array(Y)

        # Compute empirical conditional probabilities
        # Note: Can do this more efficiently...
        P = np.zeros((m, k + 1, k))
        for y in range(k):
            is_y = np.where(Y == y, 1, 0)
            for j, l in product(range(m), range(-1, k)):
                P[j, l + 1, y] = np.where(self.L[:, j] == l, 1, 0) @ is_y / is_y.sum()
        return P

    def lf_summary(
        self, Y: Optional[np.ndarray] = None, est_weights: Optional[np.ndarray] = None
    ) -> DataFrame:
        """Create a pandas DataFrame with the various per-LF statistics.

        Parameters
        ----------
        Y
            [n] or [n, 1] np.ndarray of gold labels. If provided, the empirical weight
            for each LF will be calculated.
        est_weights
            Learned weights for each LF

        Returns
        -------
        pandas.DataFrame
            Summary statistics for each LF
        """
        n, m = self.L.shape
        lf_names: Union[List[str], List[int]]
        d: OrderedDict[str, Series] = OrderedDict()
        if self._lf_names is not None:
            d["j"] = list(range(m))
            lf_names = self._lf_names
        else:
            lf_names = list(range(m))

        # Default LF stats
        d["Polarity"] = Series(data=self.lf_polarities(), index=lf_names)
        d["Coverage"] = Series(data=self.lf_coverages(), index=lf_names)
        d["Overlaps"] = Series(data=self.lf_overlaps(), index=lf_names)
        d["Conflicts"] = Series(data=self.lf_conflicts(), index=lf_names)

        if Y is not None:
            labels = np.unique(
                np.concatenate((Y.flatten(), self.L.flatten(), np.array([-1])))
            )
            confusions = [
                confusion_matrix(Y, self.L[:, i], labels=labels)[1:, 1:]
                for i in range(m)
            ]
            corrects = [np.diagonal(conf).sum() for conf in confusions]
            incorrects = [
                conf.sum() - correct for conf, correct in zip(confusions, corrects)
            ]
            accs = self.lf_empirical_accuracies(Y)
            d["Correct"] = Series(data=corrects, index=lf_names)
            d["Incorrect"] = Series(data=incorrects, index=lf_names)
            d["Emp. Acc."] = Series(data=accs, index=lf_names)

        if est_weights is not None:
            d["Learned Weight"] = Series(est_weights, index=lf_names)

        return DataFrame(data=d, index=lf_names)



import logging
import os
import re
import glob
import requests
import subprocess
import pandas as pd
from bs4 import BeautifulSoup
from datetime import datetime, timedelta
from ekorpkit.io.fetch.base import BaseFetcher
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BOKMniutes(BaseFetcher):
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        super().__init__(**args)
        self.from_date = str(self.args.scrap.from_date)
        self.base_url = self.args.scrap.base_url
        self.url = self.args.scrap.url
        self.raw_hwp_dir = self.args.scrap.raw_hwp_dir
        self.raw_txt_dir = self.args.scrap.raw_txt_dir
        os.makedirs(self.raw_hwp_dir, exist_ok=True)
        os.makedirs(self.raw_txt_dir, exist_ok=True)
        self.file_prefix = self.args.scrap.file_prefix

        user_agent = "Mozilla/5.0"
        self.headers = {"User-Agent": user_agent}
        self.minutes_urls = []
        self.minutes_hwp_files = []
        self.minutes_txt_files = []

        if self.auto.load:
            self.fetch()

    def _fetch(self):
        self.get_minutes_list()
        self.get_minutes_files()
        self.convert_hwp_to_txt()
        self.build_minutes()

    def build_minutes(self):

        docs = []
        for filepath in glob.glob(os.path.join(self.raw_txt_dir, "*.txt")):
            doc = preprocess_minutes(filepath, self.file_prefix)
            docs += doc

        data = pd.DataFrame(
            docs, columns=["id", "filename", "mdate", "rdate", "section", "text"]
        )
        data = data.dropna()

        if not self.force.download:
            if os.path.isfile(self.output_file):
                log.info(
                    "minutes file already exists. combining with the existing file.."
                )
                minutes_df = pd.read_csv(self.output_file, index_col=None)
                data = minutes_df.combine_first(data)
                data = data.drop_duplicates(subset=["id"])

        eKonf.save_data(data, self.output_file, verbose=self.verbose)
        if self.verbose:
            print(data.tail())
        log.info(f"Saved {len(docs)} documents to {self.output_file}")

    def convert_hwp_to_txt(self):
        for filepath in glob.glob(os.path.join(self.raw_hwp_dir, "*.hwp")):
            filename = os.path.basename(filepath)
            hwp_file = f"{self.raw_hwp_dir}/{filename}"
            txt_file = f"{self.raw_txt_dir}/{filename[:-4]}.txt"
            self.minutes_txt_files.append(txt_file)

            if os.path.exists(txt_file):
                log.info(f"{txt_file} already exists")
                continue

            command = f"hwp5txt --output={txt_file} {hwp_file}"
            subprocess.Popen(command, shell=True)
            log.info(f"converted {hwp_file} to {txt_file}")

    def get_minutes_files(self):
        for info in self.minutes_urls:
            self.get_minutes_file(
                info["url"], info["meeting_date"], info["release_date"]
            )

    def get_minutes_file(self, page_addr, mdate, rdate):
        filename = (
            self.file_prefix
            + "_"
            + mdate.strftime("%Y%m%d")
            + "_"
            + rdate.strftime("%Y%m%d")
            + ".hwp"
        )
        filepath = f"{self.raw_hwp_dir}/{filename}"

        if os.path.exists(filepath):
            self.minutes_hwp_files.append(filepath)
            print(f"{filepath} already exists")
            return

        page = requests.get(page_addr)
        soup = BeautifulSoup(page.content, "html.parser")
        links = soup.find("div", class_="addfile").find_all("a")
        for link in links:
            link_filename = link.get_text()
            link_filename = (
                link_filename.replace("\r", "").replace("\t", "").replace("\n", "")
            )
            if link_filename[-3:] == "hwp":
                file_addr = self.base_url + link["href"]
                file_res = requests.get(file_addr)

                with open(filepath, "wb") as f:
                    f.write(file_res.content)
                self.minutes_hwp_files.append(filepath)
                print(f"saved {link_filename} as {filename} from {file_addr}")

    # Because this is for the recent minutes, changed the address of the list to that of the rss feed of the BOK minutes.
    def get_minutes_list(self):
        from_date = datetime.strptime(self.from_date, "%Y%m%d")
        url = self.url
        page = requests.get(url, headers=self.headers)

        soup = BeautifulSoup(page.content, "html.parser")
        brdList = soup.find_all("item")

        for post in brdList:
            pubdate = post.find("pubdate").get_text().strip()
            guid = post.find("guid").get_text().strip()
            title = post.find("title").get_text().strip()
            # description = post.find("description").get_text().strip()
            # if description.replace(' ','').find('통화정책방향') >= 0:
            mdate = title[title.find(")(") + 2 : -1]
            if mdate[-1] == ".":
                mdate = mdate[:-1]
            mdate = datetime.strptime(mdate, "%Y.%m.%d")
            if mdate < from_date:
                break

            rdate = pubdate[5:16]
            rdate = datetime.strptime(rdate, "%d %b %Y")
            info = {
                "meeting_date": mdate,
                "release_date": rdate,
                "url": guid,
                "title": title,
            }
            self.minutes_urls.append(info)


def tidy_sentences(section):
    sentence_enders = re.compile(r"((?<=[함음됨임봄짐움])(\s*\n|\.|;)|(?<=다)\.)\s*")
    splits = list((m.start(), m.end()) for m in re.finditer(sentence_enders, section))
    starts = [0] + [i[1] for i in splits]
    ends = [i[0] for i in splits]
    sentences = [section[start:end] for start, end in zip(starts[:-1], ends)]
    for i, s in enumerate(sentences):
        sentences[i] = (s.replace("\n", " ").replace(" ", " ")) + "."

    text = "\n".join(sentences) if len(sentences) > 0 else ""
    return sentences, text


def preprocess_minutes(filepath, file_prefix):
    filename = os.path.basename(filepath).replace(".txt", "")
    fileinfo = filename.split("_")
    mdate = datetime.strptime(fileinfo[1], "%Y%m%d") + timedelta(hours=10)
    rdate = datetime.strptime(fileinfo[2], "%Y%m%d") + timedelta(hours=16)
    fileinfo[0] = file_prefix
    filename = "_".join(fileinfo)
    doc = []

    print("open file: {}".format(filepath))
    minutes = open(filepath, encoding="utf-8").read()

    pos = re.search(
        "(.?국내외\s?경제\s?동향.?과 관련하여,?|\(가\).+경제전망.*|\(가\) 국내외 경제동향 및 평가)\n?\s*일부 위원은",
        minutes,
        re.MULTILINE,
    )
    s1 = pos.start() if pos else -1
    # pos = re.search('(.?외환.?국제금융\s?동향.?과 관련하여,?|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*일부 위원은', minutes, re.MULTILINE)
    # pos = re.search('(.?외환.?국제금융\s?동향.?과 관련하여.*|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*일부 위원은', minutes, re.MULTILINE)
    pos = re.search(
        "(.?외환.?국제금융\s?동향.?과 관련하여.*|\(나\) 외환.국제금융\s?(및 금융시장)?\s?동향)\n?\s*(일부 위원은|대부분의 위원들은)",
        minutes,
        re.MULTILINE,
    )
    s2 = pos.start() if pos else -1
    pos = re.search(
        "(.?금융시장\s?동향.?과 관련하여,?|\(다\) 금융시장\s?동향)\n?\s*일부 위원은", minutes, re.MULTILINE
    )
    s3 = pos.start() if pos else -1
    # pos = re.search('((\((다|라)\) )?.?통화정책방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책방향.*에 관해 다음과 같은 토론이 있었음.*)\n?', minutes, re.MULTILINE)
    # pos = re.search('((\((다|라)\) )?.?통화정책.?방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책방향.*에.*토론.*)\n?', minutes, re.MULTILINE)
    # pos = re.search('((\((다|라)\) )?.?통화정책\s?방향.?에 관한 토론,?|이상과 같은 의견교환을 바탕으로.*통화정책\s?방향.*에.*토론.*)\n?', minutes, re.MULTILINE)
    pos = re.search(
        "((\((다|라)\) )?.?통화정책\s?방향.?에 관한 토론,?|이상과 같은 의견\s?교환을 바탕으로.*통화정책\s?방향.*에.*토론.*)\n?",
        minutes,
        re.MULTILINE,
    )
    s4 = pos.start() if pos else -1
    pos = re.search("(\(4\) 정부측 열석자 발언.*)\n?", minutes, re.MULTILINE)
    s5 = pos.start() if pos else -1
    pos = re.search(
        "(\(.*\) 한국은행 기준금리 결정에 관한 위원별 의견\s?개진|이상과 같은 토론에 이어 .* 관한 위원별 의견개진이 있었음.*)\n?",
        minutes,
        re.MULTILINE,
    )
    s6 = pos.start() if pos else -1
    # pos = re.search('(\(\s?.*\s?\) ()(심의결과|토의결론))\n?', minutes, re.MULTILINE)
    # s7 = pos.start() if pos else -1
    positer = re.finditer("(\(\s?.*\s?\) ()(심의결과|토의결론))\n?", minutes, re.MULTILINE)
    s7 = [pos.start() for pos in positer if pos.start() > s6]
    s7 = s7[0] if s7 else -1

    # 국내외 경제동향
    bos = s1
    eos = s2
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section1, section1_txt = tidy_sentences(section)

    # 외환․국제금융 동향
    bos = s2
    eos = s3 if s3 >= 0 else s4
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section2, section2_txt = tidy_sentences(section)

    # 금융시장 동향
    bos = s3
    eos = s4
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section3, section3_txt = tidy_sentences(section)

    # 통화정책방향
    bos = s4
    eos = s5 if s5 >= 0 else s6 if s6 >= 0 else s7
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section4, section4_txt = tidy_sentences(section)

    # 위원별 의견 개진
    bos = s6
    eos = s7
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("(일부|대부분의) 위원들?은", section, re.MULTILINE)
    bos = pos.start() if pos else -1
    section = section[bos:] if bos >= 0 else section
    section5, section5_txt = tidy_sentences(section)

    # 정부측 열석자 발언
    bos = s5
    eos = s6
    section = minutes[bos:eos] if bos >= 0 or eos >= 0 else ""
    pos = re.search("정부측 열석자 발언", section, re.MULTILINE)
    bos = pos.end() + 1 if pos else -1
    section = section[bos:] if bos >= 0 else section
    section6, section6_txt = tidy_sentences(section)

    sections = [
        "Economic Situation",
        "Foreign Currency",
        "Financial Markets",
        "Monetary Policy",
        "Participants’ Views",
        "Government’s View",
    ]
    # section_texts = (section1, section2, section3, section4, section5, section6)
    section_texts = (
        section1_txt,
        section2_txt,
        section3_txt,
        section4_txt,
        section5_txt,
        section6_txt,
    )

    print(" ==> text processing completed: {}".format(filename))

    if any(section_texts):
        for s, (section, text) in enumerate(zip(sections, section_texts)):
            # for p, text in enumerate(sentences):
            id = f"{filename}_S{s+1}"
            # id = f'{filename}_S{s+1}_P{p+1}'
            # row = (id, filename, mdate, rdate, section, p, text)
            row = (id, filename, mdate, rdate, section, text)
            doc.append(row)
    else:
        print("Empty!")

    return doc



from .checkpointer import Checkpointer, CheckpointerConfig  # noqa: F401
from .log_manager import LogManager, LogManagerConfig  # noqa: F401
from .log_writer import LogWriter, LogWriterConfig  # noqa: F401
from .tensorboard_writer import TensorBoardWriter  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Korean Sentence Splitter
# Split Korean text into sentences using heuristic algorithm.
#
# Copyright (C) 2021 Hyun-Woong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from .base import Eojeol


class MorphExtractor(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.parse_args(**kwargs)

    def parse_args(self, **kwargs):
        self.kwargs = kwargs
        self.backend = kwargs.get("backend", "mecab-python3")
        self.dicdir = kwargs.get("dicdir", None)
        self.userdic_path = kwargs.get("userdic_path", None)

    def pos(self, text):
        if self.backend in ["mecab-python3", "fugashi"]:
            from ekorpkit.models.tokenizer.mecab import MeCab

            try:
                self.mecab = MeCab(
                    dicdir=self.dicdir,
                    userdic_path=self.userdic_path,
                    backend=self.backend,
                )
            except ImportError:
                raise ImportError(
                    "\n"
                    "You must install [`fugashi` or `mecab-python3`] and `mecab_ko_dic` if you want to use `fugashi` backend.\n"
                )
        else:
            raise AttributeError(
                "Wrong backend ! currently, we only support `fugashi`, `mecab-python3`, `none` backend."
            )
        return [
            Eojeol(eojeol, pos[1]) for pos in self.mecab.pos(text) for eojeol in pos[0]
        ]




class BaseTokenizer(object):
    def __init__(self, *args, **kwargs):
        pass

    def tokenize(self, text):
        raise NotImplementedError

    def encode(self, text):
        raise NotImplementedError

    def decode(self, tokens):
        raise NotImplementedError

    def __call__(self, text):
        return self.tokenize(text)

    def __repr__(self):
        return self.__class__.__name__



from typing import List

import pandas as pd
from tqdm import tqdm

from .core import BaseTFApplier


class PandasTFApplier(BaseTFApplier):
    """TF applier for a Pandas DataFrame.

    Data points are stored as Series in a DataFrame. The TFs
    run on data points obtained via a ``pandas.DataFrame.iterrows``
    call, which is single-process and can be slow for large DataFrames.
    For large datasets, consider ``DaskTFApplier`` or ``SparkTFApplier``.
    """

    def apply_generator(self, df: pd.DataFrame, batch_size: int) -> pd.DataFrame:
        """Augment a Pandas DataFrame of data points using TFs and policy in batches.

        This method acts as a generator, yielding augmented data points for
        a given input batch of data points. This can be useful in a training
        loop when it is too memory-intensive to pregenerate all transformed
        examples.

        Parameters
        ----------
        df
            Pandas DataFrame containing data points to be transformed
        batch_size
            Batch size for generator. Yields augmented data points
            for the next ``batch_size`` input data points.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame of data points in augmented data set
        """
        batch_transformed: List[pd.Series] = []
        for i, (_, x) in enumerate(df.iterrows()):
            batch_transformed.extend(self._apply_policy_to_data_point(x))
            if (i + 1) % batch_size == 0:
                yield pd.concat(batch_transformed, axis=1).T.infer_objects()
                batch_transformed = []
        yield pd.concat(batch_transformed, axis=1).T.infer_objects()

    def apply(self, df: pd.DataFrame, progress_bar: bool = True) -> pd.DataFrame:
        """Augment a Pandas DataFrame of data points using TFs and policy.

        Parameters
        ----------
        df
            Pandas DataFrame containing data points to be transformed
        progress_bar
            Display a progress bar?

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame of data points in augmented data set
        """
        x_transformed: List[pd.Series] = []
        for _, x in tqdm(df.iterrows(), total=len(df), disable=(not progress_bar)):
            x_transformed.extend(self._apply_policy_to_data_point(x))
        return pd.concat(x_transformed, axis=1).T.infer_objects()



import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime

# Import parent class
from .base import FOMC


class Minutes(FOMC):
    """
    A convenient class for extracting minutes from the FOMC website
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        r = requests.get(self.calendar_url)
        soup = BeautifulSoup(r.text, "html.parser")

        # Getting links from current page. Meetin scripts are not available.
        if self.verbose:
            print("Getting links for minutes...")
        contents = soup.find_all(
            "a", href=re.compile("^/monetarypolicy/fomcminutes\d{8}.htm")
        )

        self.links = [content.attrs["href"] for content in contents]
        self.speakers = [
            self._speaker_from_date(self._date_from_link(x)) for x in self.links
        ]
        self.titles = ["FOMC Meeting Minutes"] * len(self.links)
        self.dates = [
            datetime.strptime(self._date_from_link(x), "%Y-%m-%d") for x in self.links
        ]
        if self.verbose:
            print("{} links found in the current page.".format(len(self.links)))

        # Archived before 2015
        if from_year <= 2014:
            print("Getting links from archive pages...")
            for year in range(from_year, 2015):
                yearly_contents = []
                fomc_yearly_url = (
                    self.base_url
                    + "/monetarypolicy/fomchistorical"
                    + str(year)
                    + ".htm"
                )
                r_year = requests.get(fomc_yearly_url)
                soup_yearly = BeautifulSoup(r_year.text, "html.parser")
                yearly_contents = soup_yearly.find_all(
                    "a",
                    href=re.compile(
                        "(^/monetarypolicy/fomcminutes|^/fomc/minutes|^/fomc/MINUTES)"
                    ),
                )
                for yearly_content in yearly_contents:
                    self.links.append(yearly_content.attrs["href"])
                    self.speakers.append(
                        self._speaker_from_date(
                            self._date_from_link(yearly_content.attrs["href"])
                        )
                    )
                    self.titles.append("FOMC Meeting Minutes")
                    self.dates.append(
                        datetime.strptime(
                            self._date_from_link(yearly_content.attrs["href"]),
                            "%Y-%m-%d",
                        )
                    )
                    # Sometimes minutes carries the first day of the meeting before 2000, so update them to the 2nd day
                    if self.dates[-1] == datetime(1996, 1, 30):
                        self.dates[-1] = datetime(1996, 1, 31)
                    elif self.dates[-1] == datetime(1996, 7, 2):
                        self.dates[-1] = datetime(1996, 7, 3)
                    elif self.dates[-1] == datetime(1997, 2, 4):
                        self.dates[-1] = datetime(1997, 2, 5)
                    elif self.dates[-1] == datetime(1997, 7, 1):
                        self.dates[-1] = datetime(1997, 7, 2)
                    elif self.dates[-1] == datetime(1998, 2, 3):
                        self.dates[-1] = datetime(1998, 2, 4)
                    elif self.dates[-1] == datetime(1998, 6, 30):
                        self.dates[-1] = datetime(1998, 7, 1)
                    elif self.dates[-1] == datetime(1999, 2, 2):
                        self.dates[-1] = datetime(1999, 2, 3)
                    elif self.dates[-1] == datetime(1999, 6, 29):
                        self.dates[-1] = datetime(1999, 6, 30)

                if self.verbose:
                    print(
                        "YEAR: {} - {} links found.".format(year, len(yearly_contents))
                    )
        print("There are total ", len(self.links), " links for ", self.content_type)

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent prcessing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        res = requests.get(self.base_url + link)
        html = res.text

        # p tag is not properly closed in many cases
        html = html.replace("<P", "<p").replace("</P>", "</p>")
        html = html.replace("<p", "</p><p").replace("</p><p", "<p", 1)

        # remove all after appendix or references
        x = re.search(
            r"(<b>references|<b>appendix|<strong>references|<strong>appendix)",
            html.lower(),
        )
        if x:
            html = html[: x.start()]
            html += "</body></html>"
        # Parse html text by BeautifulSoup
        article = BeautifulSoup(html, "html.parser")

        # if link == '/fomc/MINUTES/1994/19940517min.htm':
        #    print(article)

        # Remove footnote
        for fn in article.find_all("a", {"name": re.compile("fn\d")}):
            # if fn.parent:
            #     fn.parent.decompose()
            # else:
            #     fn.decompose()
            fn.decompose()
        # Get all p tag
        paragraphs = article.findAll("p")
        para_text = []
        for paragraph in paragraphs:
            text = re.sub(r"\s+", " ", paragraph.get_text().strip())
            if len(text) > 0 and text not in self.skip_text_list:
                para_text.append(text)
        self.articles[index] = self.segment_separator.join(para_text)



import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime

# Import parent class
from .base import FOMC


class BeigeBook(FOMC):
    """
    A convenient class for extracting statement from the FOMC website
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)
        self.beige_book_url = (
            "https://www.federalreserve.gov/monetarypolicy/beige-book-default.htm"
        )
        # Please refer to https://www.federalreserve.gov/monetarypolicy/beige-book-archive.htm for the start year

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
        from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        r = requests.get(self.beige_book_url)
        soup = BeautifulSoup(r.text, "html.parser")

        # Getting links from current page. Meetin scripts are not available.
        if self.verbose:
            print("Getting links for Beigebooks...")
        contents = soup.find_all(
            "a", href=re.compile("^/monetarypolicy/beigebook\d{6}.htm")
        )
        dates = soup.find_all(
            "a",
            href=re.compile("/monetarypolicy/files/BeigeBook_\d{8}.pdf"),
        )
        dates = [content.attrs["href"] for content in dates]
        self.links = [content.attrs["href"] for content in contents]
        self.speakers = [
            self._speaker_from_date(self._date_from_link(x)) for x in dates
        ]
        self.titles = ["FOMC Beige Book"] * len(self.links)
        self.dates = [
            datetime.strptime(self._date_from_link(x), "%Y-%m-%d") for x in dates
        ]
        # TODO: _date_from_link deos not work for beigebooks, find date from content instead
        # Correct some date in the link does not match with the meeting date
        for i, m_date in enumerate(self.dates):
            if m_date == datetime(2019, 10, 11):
                self.dates[i] = datetime(2019, 10, 4)

        if self.verbose:
            print(
                "{} links and {} dates found in the current page.".format(
                    len(self.links), len(self.dates)
                )
            )
        if from_year <= 1995:
            print("Archive only from 1996, so setting from_year as 1996...")
            from_year = 1996

        to_year = int(datetime.today().strftime("%Y"))
        if from_year <= to_year:
            print("Getting links from archive pages...")
            for year in range(from_year, to_year + 1):
                yearly_contents = []
                beige_book_annual_url = (
                    self.base_url + "/monetarypolicy/beigebook" + str(year) + ".htm"
                )
                r_year = requests.get(beige_book_annual_url)
                soup_yearly = BeautifulSoup(r_year.text, "html.parser")
                yearly_contents = soup_yearly.find_all("a", text="HTML")
                yearly_dates = soup.find_all("a", text="PDF")
                yearly_dates = [content.attrs["href"] for content in yearly_dates]
                for yearly_content, yearly_date in zip(yearly_contents, yearly_dates):
                    self.links.append(yearly_content.attrs["href"])
                    self.speakers.append(
                        self._speaker_from_date(self._date_from_link(yearly_date))
                    )
                    self.titles.append("FOMC Beige Book")
                    self.dates.append(
                        datetime.strptime(self._date_from_link(yearly_date), "%Y-%m-%d")
                    )
                    # Correct some date in the link does not match with the meeting date
                    if self.dates[-1] == datetime(2007, 6, 18):
                        self.dates[-1] = datetime(2007, 6, 28)
                    elif self.dates[-1] == datetime(2007, 8, 17):
                        self.dates[-1] = datetime(2007, 8, 16)
                    elif self.dates[-1] == datetime(2008, 1, 22):
                        self.dates[-1] = datetime(2008, 1, 21)
                    elif self.dates[-1] == datetime(2008, 3, 11):
                        self.dates[-1] = datetime(2008, 3, 10)
                    elif self.dates[-1] == datetime(2008, 10, 8):
                        self.dates[-1] = datetime(2008, 10, 7)
                    if self.verbose:
                        print(f'{self.dates[-1]}: {self.links[-1]}')

                if self.verbose:
                    print(
                        "YEAR: {} - {} links found.".format(year, len(yearly_contents))
                    )
        print("There are total ", len(self.links), " links for ", self.content_type)

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent processing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        link = self.base_url + link.replace(self.base_url, "")
        print(link)
        res = requests.get(link)
        html = res.text
        article = BeautifulSoup(html, "html.parser")
        paragraphs = article.findAll("p")
        para_text = []
        for paragraph in paragraphs:
            text = re.sub(r"\s+", " ", paragraph.get_text().strip())
            if len(text) > 0 and text not in self.skip_text_list:
                para_text.append(text)
        self.articles[index] = self.segment_separator.join(para_text)



import logging
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from pathlib import Path
from .base import _configure_font

logging.getLogger("matplotlib").setLevel(logging.WARNING)


def generate_wordclouds(
    wordclouds_args,
    fig_output_dir,
    fig_filename_format,
    title_fontsize=20,
    title_color="green",
    ncols=5,
    nrows=1,
    dpi=300,
    figsize=(20, 20),
    save=True,
    save_each=False,
    save_masked=False,
    mask_dir=None,
    fontpath=None,
    colormap="PuBu",
    verbose=True,
    **kwargs,
):
    """Wrapper function that generates wordclouds
    ** Inputs **

    ** Returns **
    wordclouds as plots
    """

    fontname, fontpath = _configure_font(fontpath=fontpath)
    plt.rcParams["font.family"] = fontname

    # for individual masked wordclouds
    masked_wc_file_format = fig_filename_format + "_{}_masked.png"
    for k, wc_args in wordclouds_args.items():
        mask_file = wc_args.get("mask_file", None)
        if mask_file and mask_dir:
            if verbose:
                print(f"Creating masked wordcloud #{k}")
            wc_args["fontpath"] = fontpath
            wc_args["dpi"] = dpi
            wc_args["colormap"] = colormap

            wc_file = masked_wc_file_format.format(k)
            fig_filepath = Path(fig_output_dir) / wc_file
            wc_args["fig_filepath"] = fig_filepath
            wc_args["mask_path"] = f"{mask_dir}/{mask_file}"
            wc_args["save"] = True
            create_wordcloud(**wc_args)

    if figsize is None:
        figsize = (nrows * 4, ncols * 5)
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
    cnt = 0
    p = 1
    wc_file_format = fig_filename_format + "_{}.png"
    wc_page_file_format = fig_filename_format + "_p{}.png"
    num_clouds = len(wordclouds_args)
    for k, wc_args in wordclouds_args.items():
        r, c = divmod(cnt, ncols)

        if verbose:
            print(f"Creating wordcloud #{k}")

        wc_args["fontpath"] = fontpath
        wc_args["dpi"] = dpi
        wc_args["colormap"] = colormap
        wc_args["mask_path"] = None
        wc_args["fig"] = fig
        wc_args["ax"] = axes[r, c]
        wc_args["save"] = save_each

        wc_file = wc_file_format.format(k)
        fig_filepath = Path(fig_output_dir) / wc_file
        wc_args["fig_filepath"] = fig_filepath
        create_wordcloud(**wc_args)

        axes[r, c].set_title(
            wc_args["title"], fontsize=title_fontsize, color=title_color
        )
        cnt += 1
        if cnt == nrows * ncols:
            if save:
                wc_file = wc_page_file_format.format(p)
                fig_filepath = Path(fig_output_dir) / wc_file
                save_subplots(fig, fig_filepath, transparent=True, dpi=dpi, verbose=verbose)
            if k < num_clouds - 1:
                p += 1
                cnt = 0
                fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
    if save and cnt < nrows * ncols:
        while cnt < nrows * ncols:
            r, c = divmod(cnt, ncols)
            axes[r, c].set_visible(False)
            cnt += 1
        wc_file = wc_page_file_format.format(p)
        fig_filepath = Path(fig_output_dir) / wc_file
        save_subplots(fig, fig_filepath, transparent=True, dpi=dpi, verbose=verbose)


def savefig(fig_filepath, transparent=True, dpi=300, **kwargs):
    plt.savefig(fig_filepath, transparent=transparent, dpi=dpi, **kwargs)


def save_subplots(fig, fig_filepath, transparent=True, dpi=300, verbose=True):
    plt.subplots_adjust(
        left=0.1, bottom=0.1, right=0.9, top=0.9, wspace=0.00, hspace=0.00
    )  # make the figure look better
    fig.tight_layout()
    Path(fig_filepath).parent.mkdir(parents=True, exist_ok=True)
    fig_filepath = str(fig_filepath)
    plt.savefig(fig_filepath, transparent=transparent, dpi=dpi)
    if verbose:
        print(f"Saved {fig_filepath}")


def create_wordcloud(
    word_freq,
    fig=None,
    ax=None,
    save=False,
    fig_filepath=None,
    mask_path=None,
    contour_width=0,
    contour_color="steelblue",
    dpi=300,
    figsize=(10, 10),
    facecolor="k",
    fontpath=None,
    colormap="PuBu",
    verbose=True,
    **kwargs,
):
    """Wrapper function that generates individual wordclouds

    ** Inputs **
    fig, ax: obj -> pyplot objects from subplots method
    save: bool -> If the user would like to save the images

    ** Returns **
    wordclouds as plots"""
    from wordcloud import WordCloud

    if figsize is not None and isinstance(figsize, str):
        figsize = eval(figsize)
    # if not fontpath:
    fontpath, fontpath = _configure_font(fontpath=fontpath)

    if mask_path is not None and Path(mask_path).is_file():
        if verbose:
            print(f"Using mask {mask_path}")
        mask = np.array(Image.open(mask_path))
        wc = WordCloud(
            font_path=fontpath,
            background_color="white",
            colormap=colormap,
            mask=mask,
            width=mask.shape[1],
            height=mask.shape[0],
            contour_width=contour_width,
            contour_color=contour_color,
        )

    else:
        wc = WordCloud(font_path=fontpath, background_color="white", colormap=colormap)

    img = wc.generate_from_frequencies(word_freq)
    if ax is not None:
        ax.imshow(img, interpolation="bilinear")
        ax.axis("off")
    else:
        plt.figure(figsize=figsize, facecolor=facecolor, dpi=dpi)
        plt.imshow(img, interpolation="bilinear")
        plt.tight_layout(pad=0)
        plt.axis("off")

    if save and fig_filepath:
        Path(fig_filepath).parent.mkdir(parents=True, exist_ok=True)
        fig_filepath = str(fig_filepath)
        if verbose > 5:
            print(f"Saving wordcloud to {fig_filepath}")
        if fig is not None:
            extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            plt.savefig(fig_filepath, bbox_inches=extent.expanded(1.1, 1.2), dpi=dpi)
        else:
            wc.to_file(fig_filepath)
            plt.savefig(fig_filepath)



import numpy as np
import pandas as pd
import logging
import sklearn
from scipy.special import softmax
from sklearn.model_selection import train_test_split
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class SimpleTrainer:
    __metaclass__ = ABCMeta
    Keys = eKonf.Keys

    def __init__(self, **args):
        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self.verbose = args.get("verbose", False)
        self._model_cfg = eKonf.to_dict(args.config)
        self._dataset = args.get(eKonf.Keys.DATASET)
        self._columns = args.get(eKonf.Keys.COLUMNS)
        self._train_ = args[eKonf.Keys.TRAIN]
        self._predict_ = args[eKonf.Keys.PREDICT]
        self._method_ = self.args.get(eKonf.Keys.METHOD)
        self._eval_cfg = args.model.eval

        self._path = self.args.path
        self._pred_file = args.path.pred.filepath

        self.model = None
        self.dataset = None
        self.splits = {}
        self.train_data = None
        self.dev_data = None
        self.test_data = None
        self.label_list = None
        self.labels_map = None

        eKonf.methods(self._method_, self)

    @abstractmethod
    def train(self):
        raise NotImplementedError("Must override train")

    @abstractmethod
    def _predict(self, data: list):
        raise NotImplementedError("Must override predict")

    def load_datasets(self):
        if self._dataset is None:
            log.warning("No dataset config found")
            return
        self.dataset = eKonf.instantiate(self._dataset)
        self.splits = self.dataset.splits

        self.train_data = self.dataset.train_data
        self.dev_data = self.dataset.dev_data
        self.test_data = self.dataset.test_data

        if self.verbose:
            print("Train data:")
            print(self.train_data.tail())
        if self.dev_data is not None:
            self._model_cfg["evaluate_during_training"] = True
            if self.verbose:
                print("Eval data:")
                print(self.dev_data.tail())
        else:
            self._model_cfg["evaluate_during_training"] = False
        if self.test_data is not None and self.verbose:
            print("Test data:")
            print(self.test_data.tail())

    def convert_to_train(self):
        return (
            self.rename_columns(self.train_data, self._train_),
            self.rename_columns(self.dev_data, self._train_),
            self.rename_columns(self.test_data, self._train_),
        )

    def rename_columns(self, data, columns):
        if not columns or data is None:
            log.info("No columns or data to rename")
            return data
        renames = {
            name: key
            for key, name in columns.items()
            if name and name != key and name in data.columns
        }
        log.info(f"Renaming columns: {renames}")
        if renames:
            data = data.copy().rename(columns=renames)
        if self.verbose:
            print(data.head())
        return data

    def convert_to_predict(self, data):
        input_col = self._predict_[self.Keys.INPUT]
        data_to_predict = data[input_col].tolist()
        if self.verbose:
            print(data_to_predict[:5])
        return data_to_predict

    def append_predictions(self, data, preds):
        predicted_column = self._predict_[self.Keys.PREDICTED]
        model_outputs_column = self._predict_[self.Keys.MODEL_OUTPUTS]
        data[predicted_column] = preds[self.Keys.PREDICTED]
        data[model_outputs_column] = preds[self.Keys.MODEL_OUTPUTS]
        pred_probs_column = self._predict_.get(self.Keys.PRED_PROBS)
        if pred_probs_column:
            data[pred_probs_column] = preds[self.Keys.PRED_PROBS]
        return data

    def predict(self, data, _predict_={}):
        if _predict_:
            self._predict_ = _predict_
        data_to_predict = self.convert_to_predict(data)
        preds = self._predict(data_to_predict)
        data = self.append_predictions(data, preds)
        return data

    def cross_val_predict(self, cv=5, dev_size=0.2, random_state=1235, shuffle=True):
        if not self.splits:
            self.load_datasets()
        if shuffle:
            data = self.dataset.data.sample(frac=1).reset_index(drop=True)
        else:
            data = self.dataset.data.reset_index(drop=True)

        splits = np.array_split(data, cv)
        pred_dfs = []
        for i, split in enumerate(splits):
            _data = pd.concat(splits[:i] + splits[i + 1 :])
            _train_data, _dev_data = train_test_split(
                _data, test_size=dev_size, random_state=random_state, shuffle=shuffle
            )
            log.info(f"Train data: {_train_data.shape}, Test data: {_dev_data.shape}")
            self.train_data = _train_data
            self.dev_data = _dev_data
            self.test_data = split

            self.train()
            log.info(f"Predicting split {i}")
            pred_df = self.predict(split)
            pred_dfs.append(pred_df)
        return pd.concat(pred_dfs)

    def eval(self):
        if not self.splits:
            self.load_datasets()

        if self.test_data is None:
            log.warning("No test data found")
            return

        data_to_predict = self.convert_to_predict(self.test_data)
        preds = self._predict(data_to_predict)
        self.pred_data = self.append_predictions(self.test_data, preds)
        eKonf.save_data(self.pred_data, self._pred_file)
        if self.verbose:
            print(self.pred_data.head())
        if self._eval_cfg:
            self._eval_cfg.labels = self.labels_list
            eKonf.instantiate(self._eval_cfg, data=self.pred_data)


class SimpleNER(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.ner import NERModel

        args = self.args
        if args.labels is None:
            labels = list(self.train_data[self.Keys.LABELS].unique())

        # Create a NERModel
        model = NERModel(
            args.model_type,
            args.model_uri,
            labels=labels,
            cuda_device=args.cuda_device,
            args=eKonf.to_dict(args),
        )

        # Train the model
        model.train_model(self.train_data, eval_data=self.dev_data)

        # Evaluate the model
        result, model_outputs, predictions = model.eval_model(self.test_data)

        # Check predictions
        # print(predictions[:5])
        return result, model_outputs, predictions


class SimpleMultiLabel(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.classification import MultiLabelClassificationModel

        args = self.args
        # Create a Model
        model = MultiLabelClassificationModel(
            args.model_type,
            args.model_uri,
            num_labels=args.num_labels,
            args=eKonf.to_dict(args),
        )

        # Train the model
        model.train_model(self.train_data, eval_df=self.dev_data)

        # Evaluate the model
        result, model_outputs, predictions = model.eval_model(self.test_data)

        # Check predictions
        # print(predictions[:5])
        return result, model_outputs, predictions


class SimpleClassification(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.classification import ClassificationModel

        args = self.args
        if not self.splits:
            self.load_datasets()
        train_data, dev_data, test_data = self.convert_to_train()

        self._model_cfg["labels_list"] = train_data[self.Keys.LABELS].unique().tolist()
        args.num_labels = len(self._model_cfg["labels_list"])

        # Create a NERModel
        model = ClassificationModel(
            args.model_type,
            args.model_uri,
            num_labels=args.num_labels,
            cuda_device=args.cuda_device,
            args=self._model_cfg,
        )
        self.label_list = model.args.labels_list
        self.labels_map = model.args.labels_map

        # Train the model
        model.train_model(
            train_data, eval_df=dev_data, acc=sklearn.metrics.accuracy_score
        )

        # Evaluate the model
        result, model_outputs, wrong_predictions = model.eval_model(
            test_data, acc=sklearn.metrics.accuracy_score
        )
        if self.verbose:
            print(f"Evaluation result: {result}")
            print(f"Wrong predictions: {wrong_predictions[:5]}")
            print(f"num_outputs: {len(model_outputs)}")
            print(f"num_wrong_predictions: {len(wrong_predictions)}")

    def load_model(self, model_dir=None):
        from simpletransformers.classification import ClassificationModel

        if model_dir is None:
            model_dir = self.args.config.best_model_dir

        self.model = ClassificationModel(self.args.model_type, model_dir)
        # , args=self._model_cfg
        self.labels_list = self.model.args.labels_list
        self.labels_map = self.model.args.labels_map
        log.info(f"Loaded model from {model_dir}")

    def _predict(self, data: list):
        if self.model is None:
            self.load_model()

        predictions, raw_outputs = self.model.predict(data)
        log.info(f"type of raw_outputs: {type(raw_outputs)}")
        prob_outputs = [softmax(output.flatten().tolist()) for output in raw_outputs]
        model_outputs = [
            dict(zip(self.labels_list, output))
            for output in prob_outputs
        ]
        pred_probs = [output.max() for output in prob_outputs]
        log.info(f"raw_output: {raw_outputs[0]}")
        return {
            self.Keys.PREDICTED.value: predictions,
            self.Keys.PRED_PROBS.value: pred_probs,
            self.Keys.MODEL_OUTPUTS.value: model_outputs,
        }



import os
import sys
import random
import itertools
import pandas as pd
import tomotopy as tp
from pathlib import Path
from collections import namedtuple
from datetime import datetime
from tqdm.auto import tqdm
import numpy as np
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from ekorpkit.io.load.list import load_wordlist, save_wordlist
from ekorpkit.io.file import save_dataframe, load_dataframe
from ekorpkit.visualize.wordcloud import generate_wordclouds, savefig
from ekorpkit.pipelines.pipe import apply


ModelSummary = namedtuple(
    "ModelSummary",
    [
        "train_dt",
        "filename",
        eKonf.Keys.CORPUS,
        "model_id",
        "model_type",
        "sample_ratio",
        "num_docs",
        "num_words",
        "total_vocabs",
        "used_vocabs",
        "iterations",
        "interval",
        "burn_in",
        "ll_per_word",
        "tw",
        "min_cf",
        "min_df",
        "rm_top",
        "k",
        "k1",
        "k2",
        "alpha",
        "eta",
        "seed",
        "perplexity",
        "u_mass",
        "c_uci",
        "c_npmi",
        "c_v",
    ],
    defaults=[None] * 29,
)

IDF = tp.TermWeight.IDF
ONE = tp.TermWeight.ONE
PMI = tp.TermWeight.PMI


class TopicModel:
    def __init__(
        self,
        model_name,
        model_dir,
        output_dir,
        num_workers=0,
        ngram=None,
        files=None,
        verbose=False,
        **kwargs,
    ):
        self.model_name = model_name
        self.model_dir = Path(str(model_dir))
        self.output_dir = Path(str(output_dir))
        self.num_workers = num_workers
        self.ngram = ngram
        self.files = files
        self.verbose = verbose

        self._raw_corpus = tp.utils.Corpus()
        self._raw_corpus_keys = None
        self.ngrams = None
        self.ngram_model = None
        self._ngram_docs = None
        self.stopwords = []
        self.docs = None
        self.corpus = None
        self.corpora = None
        self.sample_ratio = 1.0
        self.active_model_id = None
        self.model = None
        self.models = {}
        self.labels = []

        self.summary_file = Path(self.files.summary)
        self.summaries = []
        if self.summary_file.is_file():
            df = eKonf.load_data(self.summary_file, index_col=0)
            for row in df.itertuples():
                self.summaries.append(ModelSummary(*row[1:]))

        self.corpus_key_path = Path(self.files.corpus_key)
        self._raw_corpus_key_path = Path(self.files.raw_corpus_key)
        self.ngram_candidates_path = Path(self.files.ngram_candidates)
        self.ngram_model_path = Path(self.files.ngram_model)
        self.ngram_docs_path = Path(self.files.ngram_docs)
        self.stoplist_paths = self.files.stoplist
        if self.stoplist_paths is None:
            self.stoplist_paths = []
        else:
            if isinstance(self.stoplist_paths, str):
                self.stoplist_paths = [self.stoplist_paths]
            else:
                self.stoplist_paths = list(self.stoplist_paths)
        self.stopwords_path = Path(self.files.stopwords)
        self.default_stopwords_path = Path(self.files.default_stopwords)
        self.default_word_prior_path = Path(self.files.default_word_prior)
        self.word_prior_path = Path(self.files.word_prior)
        (self.model_dir).mkdir(exist_ok=True, parents=True)
        (self.output_dir / "figures/wc").mkdir(exist_ok=True, parents=True)
        (self.output_dir / "figures/train").mkdir(exist_ok=True, parents=True)
        (self.output_dir / "output/train").mkdir(exist_ok=True, parents=True)
        (self.output_dir / "figures/tune").mkdir(exist_ok=True, parents=True)
        (self.output_dir / "output/tune").mkdir(exist_ok=True, parents=True)
        (self.output_dir / "logs").mkdir(exist_ok=True, parents=True)

    def _load_raw_corpus(self, reload_corpus=False):
        def data_feeder(docs):
            for doc in docs:
                fd = doc.strip().split(maxsplit=1)
                timepoint = int(fd[0])
                yield fd[1], None, {"timepoint": timepoint}

        if not self._raw_corpus or reload_corpus:
            self._raw_corpus = tp.utils.Corpus(tokenizer=tp.utils.SimpleTokenizer())
            if self.corpora is None:
                raise ValueError("corpora is not set")
            with elapsed_timer() as elapsed:
                self.corpora.load()
                self.corpora.concat_corpora()
                df = self.corpora._data

                self._raw_corpus_keys = df[self.corpora._id_keys].values.tolist()
                self._raw_corpus.process(df[self.corpora._text_key].to_list())

                eKonf.save_data(
                    df[self.corpora.IDs],
                    self._raw_corpus_key_path,
                    verbose=self.verbose,
                )
                print("Elapsed time is %.2f seconds" % elapsed())

    def extract_ngrams(self):
        if self.ngrams is None:
            self._load_raw_corpus()
            assert self._raw_corpus, "Load a corpus first"
            with elapsed_timer() as elapsed:
                print("Extracting ngram candidates")
                self.ngrams = self._raw_corpus.extract_ngrams(
                    min_cf=self.ngram.min_cf,
                    min_df=self.ngram.min_df,
                    max_len=self.ngram.max_len,
                    max_cand=self.ngram.max_cand,
                    min_score=self.ngram.min_score,
                    normalized=self.ngram.normalized,
                    workers=self.num_workers,
                )
                # print(self.ngrams)
                ngram_list = [
                    {"words": ",".join(cand.words), "score": cand.score}
                    for cand in self.ngrams
                ]
                df = pd.DataFrame(ngram_list)
                eKonf.save_data(df, self.ngram_candidates_path, verbose=self.verbose)
                print("Elapsed time is %.2f seconds" % elapsed())

    def _load_ngram_docs(self, rebuild=False):
        if self.ngram_docs_path.is_file() and not rebuild:
            with elapsed_timer() as elapsed:
                print(f"Starting to load ngram documents from {self.ngram_docs_path}")
                self._raw_corpus = tp.utils.Corpus().load(self.ngram_docs_path)
                df = eKonf.load_data(self._raw_corpus_key_path)
                self._raw_corpus_keys = df[self.corpora._id_keys].values.tolist()
                # self._raw_corpus.load(self.ngram_doc_path)
                print(f"{len(self._raw_corpus)} documents are loaded.")
                print("Elapsed time is %.2f seconds" % elapsed())
        else:
            self.extract_ngrams()
            assert self.ngrams, "Load a ngrams first"
            print("Building ngram docs by concatenaing words in ngram list")
            self._raw_corpus.concat_ngrams(self.ngrams, self.ngram.delimiter)
            self._raw_corpus.save(self.ngram_docs_path)

    def _load_stopwords(self):
        self.stopwords = []
        if self.stopwords_path.is_file():
            self.stopwords = load_wordlist(self.stopwords_path, lowercase=True)
        else:
            if self.default_stopwords_path.is_file():
                self.stopwords = load_wordlist(
                    self.default_stopwords_path, lowercase=True
                )
            else:
                self.stopwords = ["."]
        save_wordlist(self.stopwords, self.stopwords_path)
        if self.verbose:
            print(f"{len(self.stopwords)} stopwords are loaded.")

        for path in self.stoplist_paths:
            if os.path.exists(path):
                stopwords = load_wordlist(path, lowercase=True)
                self.stopwords += stopwords
                save_wordlist(stopwords, path)
                if self.verbose:
                    print(f"{len(stopwords)} stopwords are loaded from {path}")

    def _load_word_prior(self):
        if self.word_prior_path.is_file():
            self.word_prior = eKonf.load(self.word_prior_path)
            print(self.word_prior)
        else:
            if self.default_word_prior_path.is_file():
                self.word_prior = eKonf.load(self.default_word_prior_path)
                print(self.word_prior)
            else:
                self.word_prior = {}
        eKonf.save(self.word_prior, self.word_prior_path)

    def load_corpus(
        self,
        sample_ratio=1.0,
        reload_corpus=False,
        min_num_words=5,
        min_word_len=2,
        rebuild=False,
        **kwargs,
    ):
        sample_ratio = sample_ratio if sample_ratio else self.sample_ratio
        if self.corpus and self.sample_ratio == sample_ratio and not reload_corpus:
            print("Corpus is already loaded w/ sample_ratio: {}".format(sample_ratio))
            return True
        else:
            print("Start loading corpus w/ sample_ratio: {}".format(sample_ratio))
        if not self._raw_corpus:
            self._load_ngram_docs(rebuild=rebuild)
        self._load_stopwords()
        assert self._raw_corpus, "Load ngram documents first"
        assert self.stopwords, "Load stopwords first"
        if sample_ratio and sample_ratio < 1.0 and sample_ratio > 0.0:
            docs = random.sample(
                self._raw_corpus, int(len(self._raw_corpus) * sample_ratio)
            )
            self.sample_ratio = sample_ratio
        else:
            docs = self._raw_corpus
            self.sample_ratio = 1.0
        self.corpus = tp.utils.Corpus()
        self.corpus_keys = []

        n_skipped = 0
        for i_doc, doc in tqdm(enumerate(docs)):
            words = [
                w for w in doc if w not in self.stopwords and len(w) >= min_word_len
            ]
            if len(words) > min_num_words:
                self.corpus.add_doc(words=words)
                self.corpus_keys.append(self._raw_corpus_keys[i_doc])
            else:
                if self.verbose > 5:
                    print(
                        f"Skipped - index={i_doc}, key={self._raw_corpus_keys[i_doc]}, words={list(words)}"
                    )
                n_skipped += 1
        print(f"Total {i_doc-n_skipped+1} documents are loaded.")
        print(f"Total {n_skipped} documents are removed from the corpus.")
        df = pd.DataFrame(self.corpus_keys, columns=self.corpora._id_keys)
        eKonf.save_data(
            df[self.corpora._id_keys], self.corpus_key_path, verbose=self.verbose
        )

    def infer_topics(
        self,
        output_dir=None,
        output_file=None,
        iterations=100,
        min_num_words=5,
        min_word_len=2,
        num_workers=0,
        use_batcher=True,
        minibatch_size=None,
        **kwargs,
    ):

        self._load_stopwords()
        assert self.stopwords, "Load stopwords first"
        assert self.model, "Model not found"
        print("Infer document out of the model")

        os.makedirs(os.path.abspath(output_dir), exist_ok=True)
        num_workers = num_workers if num_workers else 1
        text_key = self.corpora._text_key
        id_keys = self.corpora._id_keys

        df_ngram = eKonf.load_data(self.ngram_candidates_path)
        ngrams = []
        for ngram in df_ngram['words'].to_list():
            ngrams.append(ngram.split(','))
        
        simtok = SimpleTokenizer(
            stopwords=self.stopwords,
            min_word_len=min_word_len,
            min_num_words=min_num_words,
            ngrams=ngrams,
            ngram_delimiter=self.ngram.delimiter,
            verbose=self.verbose,
        )

        if self.corpora is None:
            raise ValueError("corpora is not set")
        with elapsed_timer() as elapsed:
            self.corpora.load()
            self.corpora.concat_corpora()
            df = self.corpora._data
            df.dropna(subset=[text_key], inplace=True)
            df[text_key] = apply(
                simtok.tokenize,
                df[text_key],
                description=f"tokenize",
                verbose=self.verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            df = df.dropna(subset=[text_key]).reset_index(drop=True)
            if self.verbose:
                print(df.tail())

            docs = []
            indexes_to_drop = []
            for ix in df.index:
                doc = df.loc[ix, text_key]
                mdoc = self.model.make_doc(doc)
                if mdoc:
                    docs.append(mdoc)
                else:
                    print(f"Skipped - {doc}")
                    indexes_to_drop.append(ix)
            df = df.drop(df.index[indexes_to_drop]).reset_index(drop=True)
            if self.verbose:
                print(f"{len(docs)} documents are loaded from: {len(df.index)}.")

            topic_dists, ll = self.model.infer(
                docs, workers=num_workers, iter=iterations
            )
            if self.verbose:
                print(topic_dists[-1:], ll)
                print(f"Total inferred: {len(topic_dists)}, from: {len(df.index)}")

            if len(topic_dists) == len(df.index):
                idx = range(len(topic_dists[0]))
                df_infer = pd.DataFrame(topic_dists, columns=[f"topic{i}" for i in idx])
                df_infer = pd.concat([df[id_keys], df_infer], axis=1)
                output_path = f"{output_dir}/{output_file}"
                eKonf.save_data(df_infer, output_path, verbose=self.verbose)
                print(f"Corpus is saved as {output_path}")
            else:
                print("The number of inferred is not same as the number of input.")

            print("Elapsed time is %.2f seconds" % elapsed())

    def save_document_topic_dists(self):
        assert self.model, "Model not found"

        topic_dists = []
        for doc in self.model.docs:
            topic_dists.append(doc.get_topic_dist())
        # print(topic_dists[-1:])

        if self.corpus_keys:
            df = pd.DataFrame(self.corpus_keys, columns=self.corpora._id_keys)
        elif self.corpus_key_path.is_file():
            df = eKonf.load_data(self.corpus_key_path, verbose=self.verbose)
        else:
            print("Corpus keys do not exist")
            return

        print(f"Total inferred: {len(topic_dists)}, from: {len(df.index)}")
        if len(topic_dists) == len(df.index):
            idx = range(len(topic_dists[0]))
            df_infer = pd.DataFrame(topic_dists, columns=[f"topic{i}" for i in idx])
            # print(df_infer.tail())
            df_infer = pd.concat([df[self.corpora._id_keys], df_infer], axis=1)
            print(df_infer.tail())

            filename = "{}-{}-topic_dists.csv".format(
                self.model_name, self.active_model_id
            )
            output_path = f"{self.model_dir}/{filename}"
            eKonf.save_data(df_infer, output_path, verbose=self.verbose)
            print(f"Corpus is saved as {output_path}")
        else:
            print("The number of inferred is not same as the number of input.")

    def tune_params(
        self,
        model_type="LDA",
        topics=[20],
        alphas=[0.1],
        etas=[0.01],
        sample_ratios=[0.1],
        tws=[IDF],
        min_cf=5,
        rm_top=0,
        min_df=0,
        burn_in=0,
        interval=10,
        iterations=100,
        seed=None,
        eval_coherence=True,
        save=False,
        save_full=False,
    ):
        """
        # Topics range
        topics = range(min_topics, max_topics, step_size)
        # Alpha parameter
        alphas = np.arange(0.01, 1, 0.3)
        # Beta parameter
        etas = np.arange(0.01, 1, 0.3)
        # Validation sets
        sample_ratios = [0.1, 0.5]
        """
        total_iters = (
            len(etas) * len(alphas) * len(topics) * len(tws) * len(sample_ratios)
        )
        exec_dt = datetime.now().strftime("%Y%m%d_%H%M%S")
        df_lls = None
        ys = []
        for i, (sr, k, a, e, tw) in tqdm(
            enumerate(
                itertools.product(sample_ratios, topics, alphas, etas, tws),
            ),
            total=total_iters,
        ):
            # train a model
            print(
                "sample_ratio: {}, k:{}, alpha:{}, eta:{}, tw:{}".format(
                    sr, k, a, e, str(tw)
                )
            )
            # self.load_corpus(sample_ratio=sr)
            df_ll, m_sum = self.train_model(
                model_type=model_type,
                sample_ratio=sr,
                k=k,
                tw=tw,
                alpha=a,
                eta=e,
                min_cf=min_cf,
                rm_top=rm_top,
                min_df=min_df,
                burn_in=burn_in,
                interval=interval,
                iterations=iterations,
                seed=seed,
                eval_coherence=eval_coherence,
                save=save,
                save_full=save_full,
            )
            margs = []
            if len(topics) > 1:
                margs.append("k={}".format(k))
            if len(alphas) > 1:
                margs.append("a={}".format(a))
            if len(etas) > 1:
                margs.append("e={}".format(e))
            if len(tws) > 1:
                margs.append("tw={}".format(tw))
            if len(sample_ratios) > 1:
                margs.append("sr={}".format(sr))
            y = ",".join(margs) if len(margs) > 0 else "ll_{}".format(i)
            ys.append(y)

            df_ll.rename(columns={"ll_per_word": y}, inplace=True)
            if df_lls is not None:
                df_lls = df_lls.merge(df_ll, on="iter")
            else:
                df_lls = df_ll

        out_file = "{}-{}-ll_per_word-{}.csv".format(
            self.model_name, model_type, exec_dt
        )
        out_file = str(self.model_dir / "output/tune" / out_file)
        df_lls.to_csv(out_file)
        ax = df_lls.plot(x="iter", y=ys, kind="line")
        ax.set_xlabel("Iterations")
        ax.set_ylabel("Log-likelihood per word")
        ax.invert_yaxis()
        out_file = "{}-{}-ll_per_word-{}.png".format(
            self.model_name, model_type, exec_dt
        )
        out_file = str(self.model_dir / "figures/tune" / out_file)
        savefig(out_file, transparent=False, dpi=300)

    def train_model(
        self,
        model_id=None,
        model_type="LDA",
        model_path=None,
        sample_ratio=None,
        k=None,
        k1=None,
        k2=None,
        t=1,
        tw=IDF,
        gamma=2,
        alpha=0.1,
        eta=0.01,
        phi=0.1,
        min_cf=5,
        rm_top=0,
        min_df=0,
        burn_in=0,
        interval=10,
        iterations=100,
        seed=None,
        eval_coherence=False,
        set_word_prior=False,
        save=True,
        save_full=True,
        **kwargs,
    ):

        self.load_corpus(sample_ratio=sample_ratio)
        assert self.corpus, "Load corpus first"

        exec_dt = datetime.now().strftime("%Y%m%d_%H%M%S")
        model_type = model_type.upper()
        if model_id is None:
            margs = [model_type]
            if k:
                margs.append("k{}".format(k))
            if k1:
                margs.append("k{}".format(k1))
            if k2:
                margs.append("k{}".format(k2))
            if self.sample_ratio < 1.0:
                margs.append("sr{}".format(int(self.sample_ratio * 100)))
            model_id = ".".join(margs)
        self.active_model_id = model_id

        if not seed:
            random.seed()
            seed = random.randint(0, 32767)

        if model_type == "LDA":
            mdl = tp.LDAModel(
                tw=tw,
                k=k,
                min_cf=min_cf,
                rm_top=rm_top,
                alpha=alpha,
                eta=eta,
                corpus=self.corpus,
                seed=seed,
            )
        elif model_type == "HPA":
            mdl = tp.HPAModel(
                tw=tw,
                k1=k1,
                k2=k2,
                min_cf=min_cf,
                rm_top=rm_top,
                alpha=alpha,
                eta=eta,
                corpus=self.corpus,
                seed=seed,
            )
        elif model_type == "HDP":
            mdl = tp.HDPModel(
                tw=tw,
                min_cf=min_cf,
                rm_top=rm_top,
                gamma=gamma,
                alpha=alpha,
                initial_k=k,
                corpus=self.corpus,
                seed=seed,
            )
        elif model_type == "CTM":
            mdl = tp.CTModel(
                tw=tw,
                k=k,
                min_cf=min_cf,
                rm_top=rm_top,
                eta=eta,
                smoothing_alpha=alpha,
                corpus=self.corpus,
                seed=seed,
            )
        elif model_type == "DTM":
            mdl = tp.DTModel(
                tw=tw,
                k=k,
                t=t,
                min_cf=min_cf,
                rm_top=rm_top,
                alpha_var=alpha,
                eta_var=eta,
                phi_var=phi,
                corpus=self.corpus,
                seed=seed,
            )
        else:
            print("{} is not supported".format(model_type))
            return False

        if set_word_prior:
            self._load_word_prior()
            for tno, words in self.word_prior.items():
                print(f"Set words {words} to topic #{tno} as prior.")
                for word in words:
                    mdl.set_word_prior(
                        word, [1.0 if i == int(tno) else 0.1 for i in range(k)]
                    )

        mdl.burn_in = burn_in
        mdl.train(0)
        print(
            "Num docs:",
            len(mdl.docs),
            ", Vocab size:",
            mdl.num_vocabs,
            ", Num words:",
            mdl.num_words,
        )
        print("Removed top words:", mdl.removed_top_words)

        print(
            "Training model by iterating over the corpus {} times, {} iterations at a time".format(
                iterations, interval
            )
        )
        ll_per_words = []
        for i in range(0, iterations, interval):
            mdl.train(interval)
            if model_type == "HDP":
                print(
                    "Iteration: {}\tLog-likelihood: {}\tNum. of topics: {}".format(
                        i, mdl.ll_per_word, mdl.live_k
                    )
                )
            else:
                print("Iteration: {}\tLog-likelihood: {}".format(i, mdl.ll_per_word))
            ll_per_words.append((i, mdl.ll_per_word))
        df_ll = pd.DataFrame(ll_per_words, columns=["iter", "ll_per_word"])
        out_file = "{}-{}-ll_per_word-{}.csv".format(
            self.model_name, self.active_model_id, exec_dt
        )
        out_file = str(self.output_dir / "output/train" / out_file)
        df_ll.to_csv(out_file)
        ax = df_ll.plot(x="iter", y="ll_per_word", kind="line")
        ax.set_xlabel("Iterations")
        ax.set_ylabel("Log-likelihood per word")
        ax.invert_yaxis()
        out_file = "{}-{}-ll_per_word-{}.png".format(
            self.model_name, self.active_model_id, exec_dt
        )
        out_file = str(self.output_dir / "figures/train" / out_file)
        savefig(out_file, transparent=False, dpi=300)

        mdl.summary()
        self.model = mdl
        self.models[model_id] = mdl

        coh_values = self.eval_coherence_value() if eval_coherence else {}
        # calculate coherence using preset

        original_stdout = sys.stdout
        out_file = "{}-{}-summary-{}.txt".format(
            self.model_name, self.active_model_id, exec_dt
        )
        out_file = str(self.output_dir / "logs" / out_file)
        with open(out_file, "w") as f:
            sys.stdout = f  # Change the standard output to the file.
            mdl.summary()
            if coh_values:
                print("<Topic Coherence Scores>")
                for cm, cv in coh_values.items():
                    print(f"| {cm}: {cv}")
            sys.stdout = original_stdout  # Reset the standard output.

        if model_path is None:
            model_path = "{}-{}.mdl".format(self.model_name, model_id)
            model_path = self.model_dir / model_path
        else:
            model_path = Path(model_path)
        if save:
            mdl.save(str(model_path), full=save_full)

        self.save_document_topic_dists()

        entry = ModelSummary(
            exec_dt,
            model_path.name,
            self.model_name,
            self.active_model_id,
            model_type,
            self.sample_ratio,
            len(mdl.docs),
            mdl.num_words,
            len(mdl.vocabs),
            len(mdl.used_vocabs),
            iterations,
            interval,
            mdl.burn_in,
            round(mdl.ll_per_word, 2),
            str(tw),
            min_cf,
            min_df,
            rm_top,
            mdl.k,
            k1,
            k2,
            alpha,
            eta,
            seed,
            round(mdl.perplexity),
            round(coh_values["u_mass"], 3) if "u_mass" in coh_values else None,
            round(coh_values["c_uci"], 3) if "c_uci" in coh_values else None,
            round(coh_values["c_npmi"], 3) if "c_npmi" in coh_values else None,
            round(coh_values["c_v"], 3) if "c_v" in coh_values else None,
        )
        self.summaries.append(entry)
        df = pd.DataFrame(self.summaries)
        eKonf.save_data(df, self.summary_file, index=True)
        return df_ll, entry

    def eval_coherence_value(
        self, coherence_metrics=["u_mass", "c_uci", "c_npmi", "c_v"]
    ):
        assert self.model, "Model not found"
        mdl = self.model
        coh_values = {}
        for metric in coherence_metrics:
            coh = tp.coherence.Coherence(mdl, coherence=metric)
            average_coherence = coh.get_score()
            coh_values[metric] = average_coherence
            coherence_per_topic = [coh.get_score(topic_id=k) for k in range(mdl.k)]
            print("==== Coherence : {} ====".format(metric))
            print("Average:", average_coherence, "\nPer Topic:", coherence_per_topic)
            print()
        return coh_values

    def load_model(self, model_id=None, model_file=None, reload_model=False, **kwargs):
        if model_id:
            self.active_model_id = model_id
        if self.active_model_id in self.models and not reload_model:
            print("The model is already loaded.")
            return True

        if model_file is None:
            model_file = "{}-{}.mdl".format(self.model_name, self.active_model_id)
        model_path = self.model_dir / model_file
        print("Loading a model from {}".format(model_path))
        if model_path.is_file():
            if not self.active_model_id:
                self.active_model_id = model_path.stem.split("-")[-1]
            model_type = self.active_model_id.split(".")[0]
            model_path = str(model_path)
            with elapsed_timer() as elapsed:
                if model_type == "LDA":
                    mdl = tp.LDAModel.load(model_path)
                elif model_type == "HPA":
                    mdl = tp.HPAModel.load(model_path)
                elif model_type == "HDP":
                    mdl = tp.HDPModel.load(model_path)
                else:
                    print("{} is not supported".format(model_type))
                    return False
                self.models[self.active_model_id] = mdl
                self.model = mdl
                print("Elapsed time is %.2f seconds" % elapsed())
        else:
            self.model = None
            print("Model file not found")

    def save_labels(self, names=None, **kwargs):
        if names is None:
            print("No names are given")
            return

        if not self.labels:
            self.label_topics()
        for k in names:
            self.labels[int(k)]["topic_name"] = names[k]
            if self.verbose:
                print(f"{k}: {names[k]}")
        label_file = "{}-labels.csv".format(self.active_model_id)
        label_file = self.output_dir / label_file
        df = pd.DataFrame(self.labels)
        eKonf.save_data(df, label_file, index=False, verbose=self.verbose)

    def label_topics(
        self,
        rebuild=False,
        use_pmiextractor=False,
        min_cf=10,
        min_df=5,
        max_len=5,
        max_cand=100,
        smoothing=1e-2,
        mu=0.25,
        window_size=100,
        top_n=10,
        **kwargs,
    ):

        label_file = "{}-labels.csv".format(self.active_model_id)
        label_file = self.output_dir / label_file
        if label_file.is_file() and not rebuild:
            print("loading labels from {}".format(label_file))
            df = eKonf.load_data(label_file)
            self.labels = df.to_dict("records")
        else:
            assert self.model, "Model not found"
            mdl = self.model
            if use_pmiextractor:
                # extract candidates for auto topic labeling
                print("extract candidates for auto topic labeling")
                extractor = tp.label.PMIExtractor(
                    min_cf=min_cf, min_df=min_df, max_len=max_len, max_cand=max_cand
                )
                with elapsed_timer() as elapsed:
                    cands = extractor.extract(mdl)
                    print("Elapsed time is %.2f seconds" % elapsed())
                    labeler = tp.label.FoRelevance(
                        mdl,
                        cands,
                        min_df=min_df,
                        smoothing=smoothing,
                        mu=mu,
                        window_size=window_size,
                    )
                    print("Elapsed time is %.2f seconds" % elapsed())
                self.labeler = labeler

            labels = []
            for k in range(mdl.k):
                print("== Topic #{} ==".format(k))
                name = f"Topic #{k}"
                if use_pmiextractor:
                    lbls = ",".join(
                        label
                        for label, score in labeler.get_topic_labels(k, top_n=top_n)
                    )
                    print(
                        "Labels:",
                        ", ".join(
                            label
                            for label, score in labeler.get_topic_labels(k, top_n=top_n)
                        ),
                    )
                wrds = ",".join(
                    word for word, prob in mdl.get_topic_words(k, top_n=top_n)
                )
                if use_pmiextractor:
                    label = {
                        "topic_no": k,
                        "topic_num": f"topic{k}",
                        "topic_name": name,
                        "topic_labels": lbls,
                        "topic_words": wrds,
                    }
                else:
                    label = {
                        "topic_no": k,
                        "topic_num": f"topic{k}",
                        "topic_name": name,
                        "topic_words": wrds,
                    }
                labels.append(label)
                for word, prob in mdl.get_topic_words(k, top_n=top_n):
                    print(word, prob, sep="\t")
                print()

            self.labels = labels
            df = pd.DataFrame(self.labels)
            eKonf.save_data(df, label_file, index=False, verbose=self.verbose)

    def visualize(self, **kwargs):
        import pyLDAvis

        assert self.model, "Model not found"
        mdl = self.model
        topic_term_dists = np.stack([mdl.get_topic_word_dist(k) for k in range(mdl.k)])
        doc_topic_dists = np.stack(
            [
                doc.get_topic_dist()
                for doc in mdl.docs
                if np.sum(doc.get_topic_dist()) == 1
            ]
        )
        doc_lengths = np.array(
            [len(doc.words) for doc in mdl.docs if np.sum(doc.get_topic_dist()) == 1]
        )
        vocab = list(mdl.used_vocabs)
        term_frequency = mdl.used_vocab_freq

        prepared_data = pyLDAvis.prepare(
            topic_term_dists, doc_topic_dists, doc_lengths, vocab, term_frequency
        )
        out_file = "{}-{}-ldavis.html".format(self.model_name, self.active_model_id)
        out_file = str(self.output_dir / "output" / out_file)
        pyLDAvis.save_html(prepared_data, out_file)

    def get_topic_words(self, top_n=10):
        """Wrapper function to extract topics from trained tomotopy HDP model

        ** Inputs **
        top_n: int -> top n words in topic based on frequencies

        ** Returns **
        topics: dict -> per topic, an arrays with top words and associated frequencies
        """
        assert self.model, "Model not found"
        mdl = self.model

        # Get most important topics by # of times they were assigned (i.e. counts)
        sorted_topics = [
            k
            for k, v in sorted(
                enumerate(mdl.get_count_by_topics()), key=lambda x: x[1], reverse=True
            )
        ]

        topics = dict()
        # For topics found, extract only those that are still assigned
        for k in sorted_topics:
            if type(mdl) in ["tomotopy.HPAModel", "tomotopy.HDPModel"]:
                if not mdl.is_live_topic(k):
                    continue  # remove un-assigned topics at the end (i.e. not alive)
            topic_wp = []
            for word, prob in mdl.get_topic_words(k, top_n=top_n):
                topic_wp.append((word, prob))

            topics[k] = topic_wp  # store topic word/frequency array

        return topics

    def topic_wordclouds(
        self,
        title_fontsize=20,
        title_color="green",
        top_n=100,
        ncols=5,
        nrows=1,
        dpi=300,
        figsize=(10, 10),
        save=True,
        mask_dir=None,
        wordclouds=None,
        save_each=False,
        save_masked=False,
        fontpath=None,
        colormap="PuBu",
        **kwargs,
    ):
        """Wrapper function that generates wordclouds for ALL topics of a tomotopy model
        ** Inputs **
        topic_dic: dict -> per topic, an arrays with top words and associated frequencies
        save: bool -> If the user would like to save the images

        ** Returns **
        wordclouds as plots
        """
        assert self.model, "Model not found"
        num_topics = self.model.k

        if figsize is not None and isinstance(figsize, str):
            figsize = eval(figsize)
        if mask_dir is None:
            mask_dir = str(self.output_dir / "figures/masks")
        fig_output_dir = str(self.output_dir / "figures/wc")
        fig_filename_format = "{}-{}-wc_topic".format(
            self.model_name, self.active_model_id
        )
        if wordclouds is None:
            wordclouds_args = {}
        else:
            wordclouds_args = eKonf.to_dict(wordclouds)
        for k in range(num_topics):
            topic_freq = dict(self.model.get_topic_words(k, top_n=top_n))
            if k in wordclouds_args:
                wc_args = wordclouds_args[k]
            else:
                wc_args = {}
            title = wc_args.get("title", None)
            if title is None:
                if self.labels:
                    topic_name = self.labels[k]["name"]
                    if topic_name.startswith("Topic #"):
                        topic_name = None
                else:
                    topic_name = None
                if topic_name:
                    title = f"Topic #{k} - {topic_name}"
                else:
                    title = f"Topic #{k}"
                wc_args["title"] = title
            wc_args["word_freq"] = topic_freq
            wordclouds_args[k] = wc_args

        generate_wordclouds(
            wordclouds_args,
            fig_output_dir,
            fig_filename_format,
            title_fontsize=title_fontsize,
            title_color=title_color,
            ncols=ncols,
            nrows=nrows,
            dpi=dpi,
            figsize=figsize,
            save=save,
            mask_dir=mask_dir,
            save_each=save_each,
            save_masked=save_masked,
            fontpath=fontpath,
            colormap=colormap,
            verbose=self.verbose,
            **kwargs,
        )


class SimpleTokenizer:
    """Class to tokenize texts for a corpus"""

    def __init__(
        self,
        stopwords=[],
        min_word_len=2,
        min_num_words=5,
        verbose=False,
        ngrams=[],
        ngram_delimiter="_",
        **kwargs,
    ):
        self.stopwords = stopwords if stopwords else []
        self.min_word_len = min_word_len
        self.min_num_words = min_num_words
        self.ngram_delimiter = ngram_delimiter
        if ngrams:
            self.ngrams = {ngram_delimiter.join(ngram): ngram for ngram in ngrams}
        else:
            self.ngrams = {}
        self.verbose = verbose
        self.verbose = verbose
        if verbose:
            print(f"{self.__class__.__name__} initialized with:")
            print(f"\tstopwords: {len(self.stopwords)}")
            print(f"\tmin_word_len: {self.min_word_len}")
            print(f"\tmin_num_words: {self.min_num_words}")
            print(f"\tngrams: {len(self.ngrams)}")
            print(f"\tngram_delimiter: {self.ngram_delimiter}")

    def tokenize(self, text):
        if text is None:
            return None
        if len(self.ngrams) > 0:
            words = text.split()
            for repl, ngram in self.ngrams.items():
                words = self.replace_seq(words, ngram, repl)
        else:
            words = text.split()
        words = [
            w for w in words if w not in self.stopwords and len(w) >= self.min_word_len
        ]
        if len(set(words)) > self.min_num_words:
            return words
        else:
            return None

    @staticmethod
    def replace_seq(sequence, subseq, repl):
        if len(sequence) < len(subseq):
            return sequence
        return eval(str(list(sequence)).replace(str(list(subseq))[1:-1], f"'{repl}'"))



import os
import logging
import pandas as pd
import numpy as np
import gensim
import gensim.downloader as api
from gensim.models import Word2Vec, KeyedVectors
from unicodedata import category
from collections import namedtuple
from tqdm.auto import tqdm
from ekorpkit.io.file import save_dataframe, load_dataframe
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class WordVector:
    __gensim__version__ = gensim.__version__

    def __init__(
        self,
        **args,
    ):
        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self.corpus = args.corpus
        self.benchmarks = args.benchmarks
        self.verbose = args.verbose
        self.wv = None

    def load(self):
        if self.args.model_type == "glove":
            self.load_glove(self.args.model_path)
        elif self.args.model_type == "word2vec":
            self.load_word2vec(self.args.model_path)
        elif self.args.model_type == "pretrained":
            self.load_pretrained(self.args.model_name)
        else:
            self.load_model(self.args.model_path)

    def load_model(self, model_file):
        self.wv = KeyedVectors.load(model_file)

    def load_word2vec(self, word2vec_file):
        self.wv = KeyedVectors.load_word2vec_format(word2vec_file)

    def load_glove(self, glove_file):
        self.wv = KeyedVectors.load_word2vec_format(
            glove_file, binary=False, no_header=True
        )

    def load_pretrained(self, model_name):
        self.wv = api.load(model_name)

    def evaluate_word_analogies(
        self,
        analogies,
        restrict_vocab=300_000,
        case_insensitive=True,
        dummy4unknown=False,
        similarity_function="most_similar",
    ):
        _analogies = self.benchmarks.analogies.get(analogies)
        if _analogies is None:
            raise ValueError(f"Analogies {analogies} not found.")

        score, sections = self.wv.evaluate_word_analogies(
            _analogies,
            restrict_vocab=restrict_vocab,
            case_insensitive=case_insensitive,
            dummy4unknown=dummy4unknown,
            similarity_function=similarity_function,
        )
        log.info(f"Evaluation score: {score}")
        summary = (
            pd.DataFrame(
                [
                    [c["section"], len(c["correct"]), len(c["incorrect"])]
                    for c in sections
                ],
                columns=["category", "correct", "incorrect"],
            )
            .assign(samples=lambda x: x.correct.add(x.incorrect))
            .assign(average=lambda x: x.correct.div(x.samples))
        )
        category_names = self.benchmarks.categories.get(analogies)
        if category_names is not None:
            category_names = eKonf.to_dict(category_names)
            summary["category"] = summary.category.replace(category_names)
        summary["corpus"] = self.corpus
        summary = summary.rename(columns=str.capitalize)

        results = {"score": score, "summary": summary}

        sections = pd.DataFrame(sections)
        for _eval in ["correct", "incorrect"]:
            res = sections[["section", _eval]].explode(_eval).reset_index(drop=True)
            res.columns = ["category", "wordpairs"]
            res["category"] = res.category.replace(category_names)
            if case_insensitive:
                res.wordpairs = res.wordpairs.map(lambda x: tuple(map(str.lower, x)))
            results[_eval] = res
        return results

    def reduce_embeddings_2d(self, restrict_vocab=100_000):

        from numpy.linalg import norm
        from sklearn.decomposition import IncrementalPCA

        vectors = self.wv.vectors[:restrict_vocab]
        vectors /= norm(vectors, axis=1).reshape(-1, 1)
        log.info(f"dimensions: {vectors.shape}")

        words = self.wv.index_to_key[:100000]
        word2idx = {w: i for i, w in enumerate(words)}

        pca = IncrementalPCA(n_components=2)

        vectors2D = pca.fit_transform(vectors)
        log.info(f"explained variance: {pca.explained_variance_ratio_}")
        return {"vectors": vectors2D, "word2idx": word2idx}

    def find_most_similar_analogies(self, eval_results, word2idx, vectors, n=1):
        from scipy.spatial.distance import cosine

        def word_to_index(wordpairs):
            idx = []
            for word in wordpairs:
                if word not in word2idx:
                    return None
                idx.append(word2idx[word])
            return tuple(idx)

        def cosine_similarity(idx):
            if idx is None:
                return None
            v1 = vectors[idx[1]] - vectors[idx[0]]
            v2 = vectors[idx[3]] - vectors[idx[2]]
            return cosine(v1, v2)

        eval_results["word2idx"] = eval_results.wordpairs.map(word_to_index)
        eval_results["similarity"] = eval_results.word2idx.map(cosine_similarity)

        # find the most similar wordpairs by cosine similarity of each category group
        eval_results = eval_results.sort_values(by="similarity", ascending=False)
        # first row of each category is the most similar wordpairs
        best_analogies = eval_results.groupby("category").first()
        return best_analogies

    def plot_similar_analogies(
        self, best_analogies, vectors, ncols=3, figsize=(15, 15)
    ):
        ax_args = eKonf.compose(config_group="visualize/plot/ax")
        axes = []
        fc = ec = "darkgrey"
        for s, (category, result) in enumerate(best_analogies.iterrows()):
            best_analogy = result.wordpairs
            analogy_idx = result.word2idx
            best_analogy = [a.capitalize() for a in best_analogy]

            coords = vectors[list(analogy_idx)].tolist()
            xlim, ylim = _get_xylims_of_wordpairs(coords)

            _ax = ax_args.copy()
            _ax.axno = s
            _ax.xlim = str(xlim)
            _ax.ylim = str(ylim)

            annotations = []
            for i in [0, 2]:
                annotations.append(
                    dict(
                        text=best_analogy[i],
                        x=coords[i + 1][0],
                        y=coords[i + 1][1],
                        xtext=coords[i][0],
                        ytext=coords[i][1],
                        arrowprops=dict(
                            width=1, headwidth=5, headlength=5, fc=fc, ec=ec, shrink=0.1
                        ),
                        fontsize=12,
                    )
                )

                annotations.append(
                    dict(
                        text=best_analogy[i + 1],
                        x=coords[i + 1][0],
                        y=coords[i + 1][1],
                        xtext=coords[i + 1][0],
                        ytext=coords[i + 1][1],
                        va="center",
                        ha="center",
                        fontsize=12,
                        color="darkred" if i == 2 else "k",
                    )
                )
            _ax.annotations = annotations
            _ax.title = category
            _ax.grid = True
            axes.append(_ax)

        num_analogies = len(best_analogies)
        nrows = (num_analogies + ncols - 1) // ncols

        cfg = eKonf.compose(config_group="visualize/plot")
        cfg.figure.figsize = figsize
        cfg.figure.fontsize = 10
        cfg.subplots.ncols = ncols
        cfg.subplots.nrows = nrows
        cfg.axes = axes
        eKonf.instantiate(cfg, data=None)


def _get_xylims_of_wordpairs(coordinates):
    coordinates = pd.DataFrame(coordinates)
    xlim, ylim = coordinates.agg(["min", "max"]).T.values
    xrange, yrange = (xlim[1] - xlim[0]) * 0.1, (ylim[1] - ylim[0]) * 0.1
    xlim[0], xlim[1] = xlim[0] - xrange, xlim[1] + xrange
    ylim[0], ylim[1] = ylim[0] - yrange, ylim[1] + yrange
    return tuple(xlim), tuple(ylim)



from ..map import BaseMapper, LambdaMapper, Mapper, lambda_mapper

"""Base classes for preprocessors.

A preprocessor is a data point to data point mapping in a labeling
pipeline. This allows Snorkel operations (e.g. LFs) to share common
preprocessing steps that make it easier to express labeling logic.
A simple example for text processing is concatenating the title and
body of an article. For a more complex example, see
``snorkel.preprocess.nlp.SpacyPreprocessor``.
"""

# Used for type checking only
# Note: subclassing as below trips up mypy
BasePreprocessor = BaseMapper


class Preprocessor(Mapper):
    """Base class for preprocessors.

    See ``snorkel.map.core.Mapper`` for details.
    """

    pass


class LambdaPreprocessor(LambdaMapper):
    """Convenience class for defining preprocessors from functions.

    See ``snorkel.map.core.LambdaMapper`` for details.
    """

    pass


class preprocessor(lambda_mapper):
    """Decorate functions to create preprocessors.

    See ``snorkel.map.core.lambda_mapper`` for details.

    Example
    -------
    >>> @preprocessor()
    ... def combine_text_preprocessor(x):
    ...     x.article = f"{x.title} {x.body}"
    ...     return x
    >>> from snorkel.preprocess.nlp import SpacyPreprocessor
    >>> spacy_preprocessor = SpacyPreprocessor("article", "article_parsed")

    We can now add our preprocessors to an LF.

    >>> preprocessors = [combine_text_preprocessor, spacy_preprocessor]
    >>> from snorkel.labeling.lf import labeling_function
    >>> @labeling_function(pre=preprocessors)
    ... def article_mentions_person(x):
    ...     for ent in x.article_parsed.ents:
    ...         if ent.label_ == "PERSON":
    ...             return ABSTAIN
    ...     return NEGATIVE
    """

    pass



import logging
from enum import Enum

# from inspect import getfullargspec as getargspec


log = logging.getLogger(__name__)


def _exist_ordered_overlap(list_s, list_l):
    if len(list_s) > len(list_l):
        return _exist_ordered_overlap(list_l, list_s)
    matched_first_item = False
    for s_i, s in enumerate(list_s):
        if not list_l:
            break
        matched = -1
        for l_i, l in enumerate(list_l):
            if s == l:
                matched = l_i
                if l_i == 0 or s_i == 0:
                    matched_first_item = True
                break
        list_l = list_l[matched + 1 :]
    if matched == -1:
        return False
    if not matched_first_item and len(list_l) > 0:
        return False
    return True


class _KEEP(str, Enum):
    """Split keys in configs used by Dataset."""

    HIGHEST_SCORE = "highest_score"
    HIGHEST_ABS_SCORE = "highest_abs_score"
    SHORTEST_WITH_SCORE = "shortest_with_score"
    LONGEST_WITH_SCORE = "longest_with_score"


def _remove_overlaps(ngram_pos_scores, keep: _KEEP = _KEEP.HIGHEST_SCORE):
    """Remove overlapping ngrams by score"""
    result = []
    unigram_pos_scores = []
    for nram_pos_score in ngram_pos_scores:
        ngram, pos, score = nram_pos_score
        if len(ngram) == 1:
            unigram_pos_scores.append(nram_pos_score)
            continue
        exist_overlap = False
        for _ngram_pos_score in ngram_pos_scores:
            _ngram, _pos, _score = _ngram_pos_score
            if _ngram == ngram or len(_ngram) == 1:
                continue
            if min(_pos) > max(pos) or max(_pos) < min(pos):
                continue
            if _exist_ordered_overlap(pos, _pos):
                if keep == _KEEP.HIGHEST_SCORE and score and _score:
                    if score < _score:
                        exist_overlap = True
                        break
                    elif score == _score:
                        if len(ngram) > len(_ngram):
                            exist_overlap = True
                            break
                        elif len(ngram) == len(_ngram):
                            if ngram < _ngram:
                                exist_overlap = True
                                break
                elif keep == _KEEP.HIGHEST_ABS_SCORE and score and _score:
                    if abs(score) < abs(_score):
                        exist_overlap = True
                        break
                    elif abs(score) == abs(_score):
                        if len(ngram) > len(_ngram):
                            exist_overlap = True
                            break
                        elif len(ngram) == len(_ngram):
                            if ngram < _ngram:
                                exist_overlap = True
                                break
                elif keep == _KEEP.SHORTEST_WITH_SCORE:
                    if len(ngram) > len(_ngram):
                        exist_overlap = True
                        break
                    elif len(ngram) == len(_ngram):
                        if score and _score:
                            if abs(score) < abs(_score):
                                exist_overlap = True
                                break
                        elif score:
                            exist_overlap = True
                            break
                        elif _score:
                            exist_overlap = True
                            break
                else:
                    if len(ngram) < len(_ngram):
                        exist_overlap = True
                        break
                    elif len(ngram) == len(_ngram):
                        if ngram < _ngram:
                            exist_overlap = True
                            break
        if not exist_overlap:
            result.append(nram_pos_score)

    for uniram_pos_score in unigram_pos_scores:
        unigram, pos, score = uniram_pos_score
        exist_overlap = False
        for _ngram_pos_score in result:
            _ngram, _pos, _score = _ngram_pos_score
            if _ngram == unigram or len(_ngram) == 1:
                continue
            if min(_pos) > max(pos) or max(_pos) < min(pos):
                continue
            if pos[0] in _pos:
                exist_overlap = True
                break
        if not exist_overlap:
            result.append(uniram_pos_score)

    result = sorted(result, key=lambda x: x[1][0], reverse=False)
    return result


def _prepare_ngram_tuples(
    words,
    max_n=5,
    max_window=None,
    max_skip=None,
    postag_rules=[],
    postag_delim=";",
    include_positions=False,
):
    num_words = len(words)
    ngrams = []
    for ngram_index_set in _get_ngram_indices(num_words, max_n, max_window, max_skip):
        ngram = tuple(words[i] for i in ngram_index_set)
        position = tuple(i for i in ngram_index_set)
        if _match_any_rules(ngram, postag_rules, postag_delim) or len(ngram) == 1:
            if include_positions:
                ngrams.append((ngram, position))
            else:
                ngrams.append(ngram)

    return ngrams


def _get_ngram_indices(num_words, max_n, max_window=None, max_skip=None):
    from itertools import combinations

    if max_window is None:
        max_window = max_n
    if max_skip is None:
        max_skip = 0
    if max_skip > max_n:
        max_skip = max_n

    word_positions = list(range(num_words))
    indices = set()
    for window in range(1, min(max_window, num_words) + 1):
        for starting in range(num_words - window + 1):
            position_set = word_positions[starting : starting + window]
            for n in range(max(1, window - max_skip), min(window, max_n) + 1):
                indices.update(set(combinations(position_set, n)))

    return sorted(indices)


def _get_ngram_tuple(ngram_str, ngram_delim=";"):
    ngram = ngram_str.split(ngram_delim)
    return tuple(ngram)


def _get_ngram_str(
    ngram_tuple,
    ngram_delim=";",
    postag_delim="/",
    strip_pos=True,
    postag_length=None,
    lowercase=True,
):
    surfaces = [
        _get_word(token, postag_delim, strip_pos, postag_length, lowercase)
        for token in ngram_tuple
    ]
    return ngram_delim.join(surfaces)


def _get_word(
    token, postag_delim="/", strip_pos=True, postag_length=None, lowercase=True
):
    if not isinstance(postag_delim, str):
        return token.lower() if lowercase else token
    token_pos = token.split(postag_delim)
    if strip_pos:
        return token_pos[0].lower() if lowercase else token_pos[0]
    return (
        (token.lower() if lowercase else token)
        if len(token_pos) == 1
        else (token_pos[0].lower() if lowercase else token_pos[0])
        + postag_delim
        + (token_pos[1][:postag_length] if postag_length else token_pos[1]).upper()
    )


def _match(ngram, postag_rule, postag_delim="/"):
    if isinstance(postag_rule, str):
        if not postag_rule.startswith(postag_delim):
            postag_rule = postag_delim + postag_rule
        for token in ngram:
            if postag_rule in token:
                return True
        return False
    if isinstance(postag_rule, list):
        postag_rule = tuple(postag_rule)
        if len(ngram) != len(postag_rule):
            return False
        for token, tag in zip(ngram, postag_rule):
            if not tag.startswith(postag_delim):
                tag = postag_delim + tag
            if tag not in token:
                return False
        return True


def _match_any_rules(
    ngram,
    postag_rules,
    postag_delim="/",
):
    if not postag_rules:
        return True
    for rule in postag_rules:
        if _match(ngram, rule, postag_delim):
            return True
    return False



import os
from setuptools import setup, find_packages
import versioneer


with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def get_about():
    about = {}
    basedir = os.path.abspath(os.path.dirname(__file__))
    with open(
        os.path.join(basedir, "ekorpkit", "conf", "about", "app", "default.yaml")
    ) as f:
        for line in f:
            k, v = line.split(": ")
            about[k.strip()] = v.strip()
    return about


def requirements():
    with open(
        os.path.join(os.path.dirname(__file__), "requirements.txt"), encoding="utf-8"
    ) as f:
        return f.read().splitlines()


def get_extra_requires(path, add_exhaustive=True):
    import re
    from collections import defaultdict

    with open(path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                tags.add(re.split("[<=>]", k.strip())[0])
                for t in tags:
                    extra_deps[t].add(k.strip())

        # add tag `exhaustive` at the end
        if add_exhaustive:
            extra_deps["exhaustive"] = set(vv for v in extra_deps.values() for vv in v)

    return extra_deps


about = get_about()
setup(
    name="ekorpkit",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author=about["author"],
    url="https://github.com/entelecheia/ekorpkit",
    description=about["description"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements(),
    extras_require=get_extra_requires("ekorpkit/resources/requirements-extra.yaml"),
    keywords=[],
    packages=find_packages(),
    python_requires=">=3.7",
    include_package_data=True,
    entry_points={
        "console_scripts": ["ekorpkit=ekorpkit.cli:hydra_main", "ekorpkit-run=ekorpkit.run:main"],
    },
)



import os
import re
import time
import codecs
import pandas as pd
from glob import glob
from ekorpkit import eKonf
from omegaconf.dictconfig import DictConfig
from pytablewriter import MarkdownTableWriter
from pytablewriter.style import Style
from pathlib import Path
from ekorpkit.utils.func import humanbytes, get_modified_time
import ekorpkit


def info_docs(**args):
    cfg = eKonf.to_config(args)

    info_cfg = cfg.info
    table_cfg = info_cfg.table
    info_name = info_cfg.name

    if info_cfg.base_dir is None:
        info_cfg.base_dir = Path(ekorpkit.__path__[0]).parent.as_posix()
    sample_output_dir = info_cfg.get("sample_output_dir", None)
    if sample_output_dir:
        sample_output_dir = f"{info_cfg.base_dir}/{sample_output_dir}"
        os.makedirs(sample_output_dir, exist_ok=True)

    info_paths = glob(info_cfg.info_files)
    markdown_template = codecs.decode(info_cfg.markdown_template, "unicode_escape")
    markdown_template = re.sub(r"\s[3,]", "\n\n", markdown_template)
    readme_md_template = codecs.decode(info_cfg.readme_md_template, "unicode_escape")
    readme_md_template = re.sub(r"\s[3,]", "\n\n", readme_md_template)

    infos = []
    for info_path in info_paths:
        print(f"Reading {info_path}")
        info = eKonf.load(info_path)

        data_file_dir = Path(info_path).parent.as_posix()
        info_output_path = (
            Path(info_cfg.base_dir) / info_cfg.info_archive_dir / f"{info.name}.yaml"
        )
        info_output_path.parent.mkdir(parents=True, exist_ok=True)
        info_src_path = (
            Path(info_cfg.base_dir) / info_cfg.info_source_dir / f"{info.name}.yaml"
        )

        if os.path.exists(info_src_path):
            info_src = eKonf.load(info_src_path)
            available = info_src.get("available", True)

            for key in info_cfg.update_info:
                if key in info_src:
                    info[key] = info_src[key]
            for key, val in info_cfg.aggregate_info.items():
                if isinstance(info.splits, DictConfig):
                    vals = [
                        split[val] for split in info.splits.values() if val in split
                    ]
                else:
                    vals = [split[val] for split in info.splits if val in split]
                info[key] = sum(vals)
            for key, value in info_cfg.modified_info.items():
                if isinstance(info.splits, DictConfig):
                    vals = [
                        get_modified_time(f"{data_file_dir}/{split[value]}")
                        for split in info.splits.values()
                        if value in split
                    ]
                else:
                    vals = [
                        get_modified_time(f"{data_file_dir}/{split[value]}")
                        for split in info.splits
                        if value in split
                    ]
                vals = [v for v in vals if v is not None]
                if vals:
                    info[key] = max(vals)
            info["info_updated"] = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
            # print(f'update info to {info_path}')
            eKonf.save(info, info_path)
            info_to_save = {}
            for key in info_cfg.info_list:
                if info.get(key) is not None:
                    info_to_save[key] = info[key]
            info_to_save = eKonf.to_config(info_to_save)

            if sample_output_dir:
                sample_text_path = Path(sample_output_dir) / f"{info.name}.txt"
            doc_output_path = (
                Path(info_cfg.base_dir) / info_cfg.info_output_dir / f"{info.name}.md"
            )
        else:
            available = False

        if available:
            infos.append(info_to_save)
            # print(f'saving info to {info_output_path}')
            eKonf.save(info_to_save, info_output_path)
            if sample_output_dir:
                save_sample_text(
                    data_file_dir, sample_text_path, info_cfg.sample_max_lines
                )

            src_link = f"{info_cfg.info_link_prefix}/{info.name}.yaml"
            sample_text_link = f"../sample/{info.name}.txt"
            fullname = info.get("fullname", info.name)
            # md_contents = markdown_template.format(name=info.fullname, src_path=src_path)
            if sample_output_dir:
                md_contents = markdown_template.format(
                    name=fullname, src_link=src_link, sample_text_link=sample_text_link
                )
            else:
                md_contents = markdown_template.format(name=fullname, src_link=src_link)
            doc_output_path.parent.mkdir(parents=True, exist_ok=True)
            # print(f'saving doc to {doc_output_path}')
            with open(doc_output_path, "w") as f:
                f.write(md_contents)
        else:
            if doc_output_path.exists():
                doc_output_path.unlink()
            if sample_output_dir:
                if sample_text_path.exists():
                    sample_text_path.unlink()

    builtins_path = Path(info_cfg.base_dir) / info_cfg.builtins_path
    builtins = eKonf.load(builtins_path)
    builtins[info_name] = [info.name for info in infos]
    eKonf.save(builtins, builtins_path)

    info_table = make_table(infos, table_cfg)
    readme_table_md = readme_md_template.format(info_table=info_table)
    readme_output_path = Path(info_cfg.base_dir) / info_cfg.readme_md_file
    readme_output_path.parent.mkdir(parents=True, exist_ok=True)
    # print(f'saving corpus to {corpus_output_path}')
    with open(readme_output_path, "w") as f:
        f.write(readme_table_md)

    info_fig_file = info_cfg.get("info_fig_file", None)
    if info_fig_file:
        fig_filepath = Path(info_cfg.base_dir) / info_fig_file
        make_figure(infos, fig_filepath)


def save_sample_text(data_file_dir, sample_text_path, sample_max_lines=None):
    sample_paths = glob(data_file_dir + "/sample-*.txt")
    if len(sample_paths) > 0:
        sample_path = sample_paths[0]
        with open(sample_text_path, "w") as f:
            with open(sample_path, "r") as sample_file:
                sample = sample_file.read().strip()
                sample = sample.split("\n")
                if sample_max_lines and len(sample) > sample_max_lines:
                    sample = sample[:sample_max_lines]
                f.write("\n".join(sample))


def make_figure(infos, fig_filepath):
    # import plotly.express as px
    import plotly.graph_objects as go

    values = []
    # KB = float(1024)
    # MB = float(KB ** 2)  # 1,048,576
    # GB = float(KB ** 3)  # 1,073,741,824

    total_size = sum([x.size_in_bytes for x in infos])
    total_size_en = sum([x.size_in_bytes for x in infos if x.lang == "en"])
    total_size_ko = sum([x.size_in_bytes for x in infos if x.lang == "ko"])
    for info in infos:
        name = info.name
        size = info.size_in_bytes
        quality = 1  # info.quality
        relative_weight = round(size * quality / total_size * 100, 2)
        values.append(
            [
                name,
                info.get("lang", "na"),
                "{:.2%}".format(relative_weight),
                relative_weight,
                size,
            ]
        )
    values.append(["Total", "", humanbytes(total_size, "GiB"), 0, total_size])
    values.append(["en", "Total", humanbytes(total_size, "GiB"), 0, total_size_en])
    values.append(["ko", "Total", humanbytes(total_size, "GiB"), 0, total_size_ko])

    headers = ["Name", "Language", "Relative_weight", "Weight", "Size"]
    df = pd.DataFrame(values, columns=headers)

    names = list(df.Name)
    weight = list(df.Weight)
    # sizes = [round(x / GB, 3) for x in list(df.Size)]
    langs = list(df.Language)
    # print(sizes)

    layout = go.Layout(
        paper_bgcolor="rgba(0,0,0,0)",
        # plot_bgcolor='rgba(0,0,0,0)'
    )
    fig = go.Figure(
        go.Treemap(
            labels=names,
            parents=langs,
            values=weight,
            textinfo="label+value",
            # values =  sizes,
            # textinfo = "label+value+percent parent+percent root",
            marker_colorscale="RdBu",
        ),
        layout=layout,
    )
    fig.update_layout(
        margin=dict(l=2, r=2, t=2, b=2),
    )

    fig.write_image(fig_filepath, scale=1.5)


def make_table(infos, table_cfg):

    df = pd.DataFrame(infos)
    eval_before = table_cfg.get("eval_before", None)
    if eval_before:
        for k, v in eval_before.items():
            df[k] = df.eval(v)
    sort_key = table_cfg.get("sort_key", None)
    if sort_key:
        df = df.sort_values(by=sort_key, ascending=False)
    calculate_weights_of = table_cfg.get("calculate_weights_of", None)
    if calculate_weights_of:
        for k, v in calculate_weights_of.items():
            df[k] = df[v].transform(lambda x: x / sum(x) * 100)
    make_links = table_cfg.get("make_links", None)
    if make_links:
        df[make_links.column] = df[make_links.column].apply(
            lambda x: make_links.link_format.format(x=x)
        )
    # df['weight'] = df['size_in_bytes'].transform(lambda x: x/sum(x)*100)
    dfs = [df]
    aggregate = table_cfg.get("aggregate", None)
    if aggregate:
        for agg in aggregate:
            group_by = agg["group_by"]
            sum_on = agg["sum"]
            names = agg["names"]
            if group_by is not None:
                df_grp = df.groupby(group_by)[sum_on].sum().reset_index()
                df_name = pd.DataFrame(names)
                df_grp = df_name.merge(df_grp, on=group_by)
                if calculate_weights_of:
                    for k, v in calculate_weights_of.items():
                        df_grp[k] = df_grp[v].transform(lambda x: x / sum(x) * 100)
            else:
                names.update({sum_on: int(df[sum_on].sum())})
                df_grp = pd.DataFrame([names])
            dfs.append(df_grp)

    df_table = pd.concat(dfs, axis=0)

    convert_to_humanbytes = table_cfg.get("convert_to_humanbytes", None)
    if convert_to_humanbytes:
        for col in convert_to_humanbytes:
            df_table[col] = df_table[col].transform(lambda x: humanbytes(x))
    format_columns = table_cfg.get("format_columns", None)
    if format_columns:
        for k, v in format_columns.items():
            df_table[k] = df_table[k].transform(lambda x: v.format(x=x))
    df_table.fillna("", inplace=True)

    headers = table_cfg.get("headers", None)
    if headers:
        df_table.rename(columns=headers, inplace=True)
        df_table = df_table[list(headers.values())]
    styles = table_cfg.get("styles", None)
    if styles:
        column_styles = [Style(**st) for st in styles]

    writer = MarkdownTableWriter(dataframe=df_table, column_styles=column_styles)
    writer.write_table()

    return writer.dumps()



from typing import List, Optional

import spacy

from ..types import FieldMap, HashingFunction

from .core import BasePreprocessor, Preprocessor

EN_CORE_WEB_SM = "en_core_web_sm"


class SpacyPreprocessor(Preprocessor):
    """Preprocessor that parses input text via a SpaCy model.

    A common approach to writing LFs over text is to first use
    a natural language parser to decompose the text into tokens,
    part-of-speech tags, etc. SpaCy (https://spacy.io/) is a
    popular tool for doing this. This preprocessor adds a
    SpaCy ``Doc`` object to the data point. A ``Doc`` object is
    a sequence of ``Token`` objects, which contain information
    on lemmatization, parts-of-speech, etc. ``Doc`` objects also
    contain fields like ``Doc.ents``, a list of named entities,
    and ``Doc.noun_chunks``, a list of noun phrases. For details
    of SpaCy ``Doc`` objects and a full attribute listing,
    see https://spacy.io/api/doc.

    Parameters
    ----------
    text_field
        Name of data point text field to input
    doc_field
        Name of data point field to output parsed document to
    language
        SpaCy model to load
        See https://spacy.io/usage/models#usage
    disable
        List of pipeline components to disable
        See https://spacy.io/usage/processing-pipelines#disabling
    pre
        Preprocessors to run before this preprocessor is executed
    memoize
        Memoize preprocessor outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)
    gpu
        Prefer Spacy GPU processing?
    """

    def __init__(
        self,
        text_field: str,
        doc_field: str,
        language: str = EN_CORE_WEB_SM,
        disable: Optional[List[str]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
        memoize: bool = False,
        memoize_key: Optional[HashingFunction] = None,
        gpu: bool = False,
    ) -> None:
        name = type(self).__name__
        super().__init__(
            name,
            field_names=dict(text=text_field),
            mapped_field_names=dict(doc=doc_field),
            pre=pre,
            memoize=memoize,
            memoize_key=memoize_key,
        )
        self.gpu = gpu
        if self.gpu:
            spacy.prefer_gpu()
        self._nlp = spacy.load(language, disable=disable or [])

    def run(self, text: str) -> FieldMap:  # type: ignore
        """Run the SpaCy model on input text.

        Parameters
        ----------
        text
            Text of document to parse

        Returns
        -------
        FieldMap
            Dictionary with a single key (``"doc"``), mapping to the
            parsed SpaCy ``Doc`` object
        """
        # Note: not trying to add the fields of `Doc` to top-level
        # as most are Cython property methods computed on the fly.
        return dict(doc=self._nlp(text))



import os
import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime
import pdfplumber

# Import parent class
from .base import FOMC


class PresConfScript(FOMC):
    """
    A convenient class for extracting press conference scripts from the FOMC website.
    It is only available from April 2011.
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)
        left_pct = 0.05  # % Distance of left side of character from left side of page.
        top_pct = 0.10  # % Distance of top of character from top of page.
        right_pct = (
            0.95  # % Distance of right side of character from left side of page.
        )
        bottom_pct = 0.90  #% Distance of bottom of the character from top of page.
        self.crop_coords = [left_pct, top_pct, right_pct, bottom_pct]

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        r = requests.get(self.calendar_url)
        soup = BeautifulSoup(r.text, "html.parser")

        if self.verbose:
            print("Getting links for press conference scripts...")
        presconfs = soup.find_all(
            "a", href=re.compile("^/monetarypolicy/fomcpresconf\d{8}.htm")
        )
        presconf_urls = [
            self.base_url + presconf.attrs["href"] for presconf in presconfs
        ]
        for presconf_url in presconf_urls:
            r_presconf = requests.get(presconf_url)
            soup_presconf = BeautifulSoup(r_presconf.text, "html.parser")
            contents = soup_presconf.find_all(
                "a", href=re.compile("^/mediacenter/files/FOMCpresconf\d{8}.pdf")
            )
            for content in contents:
                # print(content)
                self.links.append(content.attrs["href"])
                self.speakers.append(
                    self._speaker_from_date(self._date_from_link(content.attrs["href"]))
                )
                self.titles.append("FOMC Press Conference Transcript")
                self.dates.append(
                    datetime.strptime(
                        self._date_from_link(content.attrs["href"]), "%Y-%m-%d"
                    )
                )
        if self.verbose:
            print("{} links found in current page.".format(len(self.links)))

        # Archived before 2015
        if from_year <= 2014:
            print("Getting links from archive pages...")
            for year in range(from_year, 2015):
                yearly_contents = []
                fomc_yearly_url = (
                    self.base_url
                    + "/monetarypolicy/fomchistorical"
                    + str(year)
                    + ".htm"
                )
                r_year = requests.get(fomc_yearly_url)
                soup_yearly = BeautifulSoup(r_year.text, "html.parser")

                presconf_hists = soup_yearly.find_all(
                    "a", href=re.compile("^/monetarypolicy/fomcpresconf\d{8}.htm")
                )
                presconf_hist_urls = [
                    self.base_url + presconf_hist.attrs["href"]
                    for presconf_hist in presconf_hists
                ]
                for presconf_hist_url in presconf_hist_urls:
                    # print(presconf_hist_url)
                    r_presconf_hist = requests.get(presconf_hist_url)
                    soup_presconf_hist = BeautifulSoup(
                        r_presconf_hist.text, "html.parser"
                    )
                    yearly_contents = soup_presconf_hist.find_all(
                        "a",
                        href=re.compile("^/mediacenter/files/FOMCpresconf\d{8}.pdf"),
                    )
                    for yearly_content in yearly_contents:
                        # print(yearly_content)
                        self.links.append(yearly_content.attrs["href"])
                        self.speakers.append(
                            self._speaker_from_date(
                                self._date_from_link(yearly_content.attrs["href"])
                            )
                        )
                        self.titles.append("FOMC Press Conference Transcript")
                        self.dates.append(
                            datetime.strptime(
                                self._date_from_link(yearly_content.attrs["href"]),
                                "%Y-%m-%d",
                            )
                        )
                if self.verbose:
                    print(
                        "YEAR: {} - {} links found.".format(
                            year, len(presconf_hist_urls)
                        )
                    )
            print("There are total ", len(self.links), " links for ", self.content_type)

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent prcessing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        link_url = self.base_url + link
        pdf_filepath = (
            self.output_raw_dir
            + "/FOMC_PresConfScript_"
            + self._date_from_link(link)
            + ".pdf"
        )

        if not os.path.exists(pdf_filepath) or self.force_download:
            # Scripts are provided only in pdf. Save the pdf and pass the content
            res = requests.get(link_url)

            with open(pdf_filepath, "wb") as f:
                f.write(res.content)
        else:
            if self.verbose:
                print("File already exists: ", pdf_filepath)

        # Extract text from the pdf
        pdf_file_parsed = ""  # new line
        with pdfplumber.open(pdf_filepath) as pdf:
            for page in pdf.pages:
                pg_width = page.width
                pg_height = page.height
                pg_bbox = (
                    self.crop_coords[0] * float(pg_width),
                    self.crop_coords[1] * float(pg_height),
                    self.crop_coords[2] * float(pg_width),
                    self.crop_coords[3] * float(pg_height),
                )
                page_crop = page.crop(bbox=pg_bbox)
                text = page_crop.extract_text()
                pdf_file_parsed = pdf_file_parsed + "\n" + text
        paragraphs = re.sub("(\n)(\n)+", "\n", pdf_file_parsed.strip())
        paragraphs = paragraphs.split("\n")

        section = -1
        paragraph_sections = []
        for paragraph in paragraphs:
            if not re.search(
                "^(page|january|february|march|april|may|june|july|august|september|october|november|december|jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)",
                paragraph.lower(),
            ):
                if len(re.findall(r"[A-Z]", paragraph[:10])) > 5 and not re.search(
                    "(present|frb/us|abs cdo|libor|rp–ioer|lsaps|cusip|nairu|s cpi|clos, r)",
                    paragraph[:10].lower(),
                ):
                    section += 1
                    paragraph_sections.append("")
                if section >= 0:
                    paragraph_sections[section] += paragraph
        self.articles[index] = self.segment_separator.join(
            [paragraph for paragraph in paragraph_sections]
        )



"""Generic model analysis utilities shared across Snorkel."""

from .error_analysis import get_label_buckets, get_label_instances  # noqa: F401
from .metrics import metric_score  # noqa: F401
from .scorer import Scorer  # noqa: F401



import logging
import time
from pathlib import Path
from ekorpkit import eKonf
from ekorpkit.utils.func import humanbytes, get_modified_time
from ekorpkit.utils.func import elapsed_timer


log = logging.getLogger(__name__)


class SummaryInfo:
    def __init__(self, **args):
        self.args = eKonf.to_dict(args)
        self.name = self.args["name"]
        self.stat_args = self.args.get("stats", None)
        self.data_dir = self.args.get("data_dir", None)
        self.verbose = self.args.get("verbose", False)
        self.info_file = self.args.get("info_file", None)
        if self.info_file is None:
            self.info_file = f"info-{self.name}.yaml"
        self.info_path = Path(self.data_dir) / self.info_file
        self.info_list = self.args.get("info_list", [])
        self.stat_before_processing = self.args.get("stat_before_processing", None)
        self.update_files_info = self.args.get("update_files_info", {})
        self.aggregate_info = self.args.get("aggregate_info", {})
        self.modified_info = self.args.get("modified_info", {})

        self.info = {}
        self.stats = {}
        self.splits = {}

    def load(self, info={}):
        if self.info_path.exists():
            self.info = eKonf.to_dict(eKonf.load(self.info_path))
        else:
            self.info = {}

        for key in self.info_list:
            if info.get(key) is not None:
                self.info[key] = info[key]

        if "stats" in self.info:
            self.stats = self.info["stats"]
        if "splits" in self.info:
            self.splits = self.info["splits"]

        if self.verbose:
            log.info(f"Loading info file: {self.info_path}")
            eKonf.pprint(self.info)

    def init_stats(self, df=None, split_name=None, stats={}):
        if self.verbose:
            log.info(
                f"Initializing statistics for split: {split_name} with stats: {stats}"
            )
        if split_name:
            stats["name"] = split_name
            self.splits[split_name] = stats
        else:
            self.stats.update(stats)

        stat_args = self.stat_before_processing
        if stat_args is not None and df is not None:
            with elapsed_timer(format_time=True) as elapsed:
                for k, v in self.stat_args.items():
                    if k not in stat_args:
                        stat_args[k] = v

                stat_fn = eKonf.instantiate(stat_args)
                stats = stat_fn(df)
                # stats = summary_stats(df, **stat_args)
                log.info(
                    f" >> elapsed time to calculate statistics before processing: {elapsed()}"
                )
            stats.update(stats)

        if split_name:
            self.splits[split_name].update(stats)
            if self.verbose:
                log.info(f" >> updated splits: {self.splits}")
        else:
            self.stats.update(stats)
            if self.verbose:
                log.info(f" >> updated stats: {self.stats}")

    def calculate_stats(self, df, split_name=None):
        if self.verbose:
            log.info(f"Calculating statistics for split: {split_name}")
        if split_name and split_name not in self.splits:
            self.init_stats(df, split_name)
        with elapsed_timer(format_time=True) as elapsed:
            stat_fn = eKonf.instantiate(self.stat_args)
            stats = stat_fn(df)
            log.info(f" >> elapsed time to calculate statistics: {elapsed()}")

        if split_name:
            self.splits[split_name].update(stats)
        else:
            self.stats.update(stats)

    def _update_info(self):
        if self.stats:
            if "stats" not in self.info:
                self.info["stats"] = {}
            self.info["stats"].update(self.stats)
        else:
            if "stats" in self.info and not self.info["stats"]:
                del self.info["stats"]

        if self.splits:
            if "splits" not in self.info:
                self.info["splits"] = {}
            self.info["splits"].update(self.splits)

            files_info = {key: {} for key in self.update_files_info}
            for i, (split_name, split_info) in enumerate(self.splits.items()):
                for key, val in self.update_files_info.items():
                    if val in split_info:
                        files_info[key][split_name] = split_info[val]
                for key, val in self.aggregate_info.items():
                    if val in split_info:
                        if i == 0:
                            self.info[key] = split_info[val]
                        else:
                            self.info[key] += split_info[val]
            self.info["size_in_human_bytes"] = humanbytes(self.info["size_in_bytes"])

            for key in self.update_files_info:
                self.info[key] = files_info[key]

            for key, value in self.modified_info.items():
                vals = [
                    get_modified_time(f"{self.data_dir}/{split[value]}")
                    for split in self.info["splits"].values()
                    if value in split
                ]
                vals = [v for v in vals if v is not None]
                if vals:
                    self.info[key] = max(vals)
        else:
            if "splits" in self.info and not self.info["splits"]:
                del self.info["splits"]

        self.info["info_updated"] = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())

    def save(self, info={}):
        self._update_info()
        for key in info:
            if key in self.info_list and info.get(key) is not None:
                self.info[key] = info[key]

        eKonf.save(self.info, f=self.info_path)
        if self.verbose:
            log.info(f"Saving updated info file: {self.info_path}")
            eKonf.pprint(self.info)


def summary_stats(
    df,
    **args,
):

    args = eKonf.to_dict(args)
    num_columns = args.get("num_columns", None)
    agg_funcs = args.get("agg_funcs", None)
    rename_columns = args.get("rename_columns", None)
    convert_to_humanbytes = args.get("convert_to_humanbytes", None)
    num_workers = args.get("num_workers", None)
    _func_ = args.get(eKonf.Keys.FUNC, None)

    df = df.copy(deep=True)
    num_workers = num_workers if num_workers else 1
    text_keys = args.get("text_keys")
    if text_keys is None:
        text_keys = "text"
    if isinstance(text_keys, list):
        for col in text_keys:
            df[col].fillna("", inplace=True)
            df[col] = df[col].astype(str)
        separator = "\n\n"
        text_key = text_keys[0]
        df[text_key] = df[text_keys].agg(separator.join, axis=1)
    else:
        text_key = text_keys
        df[text_key] = df[text_key].astype(str)

    for col, _func_name_ in num_columns.items():
        len_func = eKonf.partial(_func_[_func_name_])
        df[col] = eKonf.apply(
            len_func, df[text_key], description=f"apply {_func_name_} to {col}"
        )

    agg_funcs = {k: list(v) for k, v in agg_funcs.items()}
    df_sum = df.groupby(lambda _: True).agg(agg_funcs)
    df_sum.columns = ["_".join(x) for x in df_sum.columns.values]
    df_sum.rename(columns=dict(rename_columns), inplace=True)
    info = df_sum.to_dict(orient="records")[0]
    if convert_to_humanbytes:
        for k, v in convert_to_humanbytes.items():
            if k in info:
                info[v] = humanbytes(info[k])
    return info


def feature_stats(
    df,
    **args,
):

    args = eKonf.to_dict(args)
    stats = args.get("stats", None)
    convert_to_humanbytes = args.get("convert_to_humanbytes", None)

    df = df.copy(deep=True)

    info = {}
    for name, expr in stats.items():
        info[name] = eval(expr)
    if convert_to_humanbytes:
        for k, v in convert_to_humanbytes.items():
            if k in info:
                info[v] = humanbytes(info[k])
    return info
